;;;;  active-set-method.lisp - simplex intersection and complex overlay functions.
;;;;  Copyright (C) 2010  Norbert Paul
;;;;
;;;;  This program is free software: you can redistribute it and/or modify
;;;;  it under the terms of the GNU General Public License as published by
;;;;  the Free Software Foundation, either version 3 of the License, or
;;;;  (at your option) any later version.
;;;;
;;;;  This program is distributed in the hope that it will be useful,
;;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;  GNU General Public License for more details.
;;;;
;;;;  You should have received a copy of the GNU General Public License
;;;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;;  Contact: norbert.paul@kit.edu

(eval-when (:load-toplevel :compile-toplevel)
  (load "utils.lisp")
  (load "dgesv-test.lisp"))

(declaim (optimize (debug 3)))

(defpackage :active-set-method
  (:use :common-lisp :utils)
  (:export
   #:initially-active 
   #:kkt-distance-matrix
   #:kkt-distance-matrix2
   #:simplex-intersection-vertices
   #:simplex-intersection-sides 
   #:simplex-intersection-edges
   #:simplex-intersection-faces))

(in-package :active-set-method)

(eval-when (:load-toplevel :compile-toplevel)
  (load "chains-matrices"))

(defun initially-active (n)
"Computes an n+2 square matrix of double-floats where 
the first n diagonal entries are set to 1, the rest is 
set to 0."

  (let* ((n+2 (+ n 2))
	 (result (make-array `(,n+2 ,n+2) 
			    :element-type 'double-float
			    :initial-element 0d0)))
    (dotimes (i n) 
      (setf (aref result i i) 1d0))

    result ) )

(defmacro TODO (what)
  (declare (ignorable what))
  ())


(defmacro to-dfloat-array (xpr &rest garbage)
  (declare (ignore garbage))
  (TODO "document")
  `(coerce ,xpr '(vector double-float))
)

(defun kkt-distance-matrix (simplex1 simplex2)

  "Computes the KKT-matrix for the distance function between 
simplex1 and simplex2 where no vertex is active."

(declare (type sequence ;instead of (simple-array (simple-array * 1) 1) 
	       simplex1 simplex2))

(declare (ftype (function (sequence ;(simple-array (simple-array * 1) 1) 
			   sequence ;(simple-array (simple-array * 1) 1)
			   )
			  (simple-array double-float 2))
		kkt-distance-matrix))

(declare (optimize (speed 3)))

  (let* ((n   (length simplex1))
	 (m   (length simplex2))
	 (n+m   (+ n m))
	 (n+m+1 (1+ n+m))
	 (n+m+2 (1+ n+m+1))
	 (result (make-array `(,n+m+2 ,n+m+2) 
			     :element-type 'double-float
			     :initial-element 0d0))
	 (mij 0d0))

    (declare (type fixnum n m n+m n+m+1 n+m+2)
	     (type double-float mij)
	     (type (simple-array double-float 2) result))

    (dotimes (i n) (declare (type fixnum i))
      (setf (aref result i   n+m) 1d0
	    (aref result n+m i)   1d0)
      (let 
	  ((ai (coerce (elt simplex1 i) '(vector double-float))))
	(declare (type (simple-array * 1) ai))
	(setf mij                (laroutines:ddot ai ai)
	      (aref result i i)  mij)
	(loop :for j fixnum :from (1+ i) :below n
	   :do
	   (setf mij               (laroutines:ddot ai (to-dfloat-array (elt simplex1 j)))
		 (aref result i j) mij
		 (aref result j i) mij))
	(loop :for n+j fixnum :from n :below n+m
	   :for j fixnum = 0 :then (1+ j)
	   :do
	   (setf mij            (- (laroutines:ddot ai (to-dfloat-array (elt simplex2 j))))
		 (aref result i   n+j) mij
		 (aref result n+j i)   mij))))
    (loop 
       :for n+j fixnum :from n :below n+m 
       :for j   fixnum = 0 :then (1+ j)
       :do
       (setf (aref result n+j   n+m+1) 1d0
	     (aref result n+m+1 n+j)   1d0)     
       (let ((bj (coerce (elt simplex2 j) '(vector double-float))))
	 (setf (aref result n+j n+j) (laroutines:ddot bj bj))
	 (loop :for n+j2 fixnum :from (1+ n+j) :below n+m
	    :for    j2 fixnum = (1+ j) :then (1+ j2)
	    :do
	    (setf mij              (laroutines:ddot bj (to-dfloat-array (elt simplex2 j2)))
		  (aref result n+j      n+j2) mij
		  (aref result n+j2 n+j)      mij))))
    result ))



(defun kkt-distance-matrix2 (simplex1 simplex2 
			     &optional (dot #'(lambda (a b) 
						(laroutines:ddot
						 (to-dfloat-array a)
						 (to-dfloat-array b))))
			     (invert #'-)
			     (one 1d0) (zero 0d0))

  "Computes the KKT-matrix for the distance function between 
simplex1 and simplex2 where no vertex is active.

This is an algebraic version where 1, 0, dot product, and inversion 
may be specified, too."

  (let* ((n   (length simplex1))
	 (m   (length simplex2))
	 (n+m   (+ n m))
	 (n+m+1 (1+ n+m))
	 (n+m+2 (1+ n+m+1))
	 (result (make-array `(,n+m+2 ,n+m+2) 
			     :initial-element zero))
	 (mij 0))

    (dotimes (i n) 
      (setf (aref result i   n+m) one
	    (aref result n+m i)   one)
      (let 
	  ((ai (elt simplex1 i)))
	(setf mij                (funcall dot ai ai)
	      (aref result i i)  mij)
	(loop :for j fixnum :from (1+ i) :below n
	   :do
	   (setf mij               (funcall dot ai (elt simplex1 j))
		 (aref result i j) mij
		 (aref result j i) mij))
	(loop :for n+j fixnum :from n :below n+m
	   :for j fixnum = 0 :then (1+ j)
	   :do
	   (setf mij            (funcall invert (funcall dot ai (elt simplex2 j)))
		 (aref result i   n+j) mij
		 (aref result n+j i)   mij))))
    (loop 
       :for n+j fixnum :from n :below n+m 
       :for j   fixnum = 0 :then (1+ j)
       :do
       (print j)
       (setf (aref result n+j   n+m+1) one
	     (aref result n+m+1 n+j)   one)     
       (let ((bj (elt simplex2 j)))
	 (setf (aref result n+j n+j) (funcall dot bj bj))
	 (loop :for n+j2 :from (1+ n+j) :below n+m
	    :for    j2 = (1+ j) :then (1+ j2)
	    :do
	    (print `(,j ,j2 ,n+j2))
	    (setf mij              (funcall dot bj (elt simplex2 j2))
		  (aref result n+j  n+j2) mij
		  (aref result n+j2 n+j ) mij))))
    result ))


(defun kkt-distance-right-hand-side (n+m)
  "Computes the right-hand side of the KKT system for distance computation:
A vector of double floats of length n+m+2 where the first n+m entries are zero 
 (as double-float 0d0)  and the last two entries are one (1d0).

Example: (kkt-drhs 3) = #(0d0 0d0 0d0 1d1 1d1)"
  (let ((result (make-array (+ n+m 2)
			    :element-type 'double-float
			    :initial-element 0d0)))
    (setf (aref result n+m) 1d0
	  (aref result (1+ n+m)) 1d0)
    
    result ))


(defun activate (i matrix &optional (zero 0d0) (one 1d0))

  "Modifies the specified matrix by setting row and column i 
to #(0d0 ... 1d0 ... 0d), hence to the row and column 
i of the identity matrix, and returns that matrix. 
The optional arguments zero and one may be specified 
to other additive and multiplicative neutral elements."

  (dotimes (j (first (array-dimensions matrix)))
    (setf (aref matrix i j) zero
          (aref matrix j i) zero ) )
  (setf (aref matrix i i) one) 

  matrix

)

(defun de-activate (i matrix backup)

  "Modifies the specified matrix by by copying 
row and column i from the backup matrix. 
The modified matrix is then returned.
"

  (dotimes (j (first (array-dimensions matrix)))
    (setf (aref matrix i j) (aref backup i j)
          (aref matrix j i) (aref backup j i) ) )
  matrix
)


(defun delete-consecutive-duplicates (lst &optional (test #'eql))
  "Modifies the given list by deleting the second of two consecutive 
entries in the list and then returns that list. 
Note that this deletes duplicates if the list elements are linearly 
ordered."

  (let ((result (cons 'head lst)))
    (loop :for tail = (cdr result) :then (cdr tail)
       :while (cdr tail)
       :do (loop :while (and (cdr tail) (funcall test (car tail) (cadr tail)))
	         :do    (rplacd tail (cddr tail))))
    (cdr result)))


(defun delete-consecutive-duplicates2 (lst &optional (test #'eql))
  "Modifies the given list by deleting the second of two consecutive 
entries in the list and then returns that list. 
Note that this deletes duplicates if the list elements are linearly 
ordered."

(do ((result (cons 'head lst))
     (tail   lst  (cdr tail)))
    ((null (cdr tail)) (cdr result))
  (do () 
      ((not (and (cdr tail) (funcall test (car tail) (cadr tail)))))
    (rplacd tail (cddr tail)))))


(defun set-active (lst matrix backup &optional (zero 0d0) (one 1d0))

  "Modifies matrix by setting all indexes specified in lst active setting all others 
inactive. 

As a side-effect the list lst will be sorted and duplicates will be deleted.
Equivalent to 

   (progn 
     (delete-duplicates (sort lst #'<))
     (dotimes (i (first (array-dimension matrix)))
       (deactivate i matrix backup))
     (dolist (i lst) (activate i matrix zero one))
     matrix)

but (maybe) faster."

  (let* ((slst   (delete-consecutive-duplicates (sort lst #'<)))
	 (slst-i slst)
	 (slst-j slst)
 	 (n      (first (array-dimensions matrix))))
    (when (and slst-i (minusp (first slst-i)))
      (error "A negative active index was specified."))
    (loop :for i :below n ; iterate rows
       :do 
       (if (eql (car slst-i) i)
	   (progn ; activate
	     (loop :while (eql (car slst-i) i) :do (pop slst-i))
	     (loop :for j :below n
		:do 
		(setf (aref matrix i j) zero
		      (aref matrix j i) zero))
	     (setf (aref matrix i i) one))
	   (progn ; deactivate 
	     (setf slst-j slst)
	     (loop :for j :below n :do 
		(if (eql (car slst-j) j) 
		    (loop :while (eql (car slst-j) j) :do (pop slst-j))
		    (setf (aref matrix i j) (aref backup i j)))))))
    (values matrix slst) ) )


(defun set-inactive (lst matrix backup &optional (zero 0d0) (one 1d0))
  "Sets all indexes specified in lst inactive and setting all others 
active. 

As as side effect the list lst will be sorted and duplicates will be deleted.
Equivalent to 
   (progn 
     (delete-duplicates (sort lst #'<))
     (dotimes (i (first (array-dimension matrix)))
       (activate i matrix backup))
     (dolist (i lst) (deactivate i matrix zero one))
     matrix)."

  (let* ((slst   (delete-consecutive-duplicates (sort lst #'<)))
	 (slst-i slst)
	 (slst-j slst)
 	 (n      (first (array-dimensions matrix))))
    (when (and slst-i (minusp (first slst-i)))
      (error "A negative active index was specified."))
    (loop :for i :below n ; iterate rows
       :do 
       (if (not (eql (car slst-i) i))
	   (progn ; activate
	     (loop :for j :below n
		:do 
		(setf (aref matrix i j) zero
		      (aref matrix j i) zero))
	     (setf (aref matrix i i) one))
	   (progn ; deactivate 
	     (pop slst-i)
	     (setf slst-j slst)
	     (loop :for j :below n :do 
		(when (eql (car slst-j) j) 
		  (pop slst-j)
		  (setf (aref matrix i j) (aref backup i j)))))))
    (values matrix slst) )
)

(defparameter *fuzz* (laroutines:depsilon))

(defun fuzzy-zerop (n &optional (fuzz *fuzz*))
  (<= (- fuzz) n  fuzz)
)

(defun fuzzy-minusp (n &optional (fuzz *fuzz*))
  (< n (- fuzz))
)

(defun simplex-intersection-vertices (simplex1 simplex2 &optional (epsilon *fuzz*))
  "Computes the intersection vertices of simplex1 and simplex2 and returns the 
corresponding inactive sets of the vertex indexes. 

Each vertex is specified as a sequence of simple arrays, the vertices.
Each such vertex array is a simple array of vertex coordinates which must be 
coercible to double-float.

Inactive sets are minimal sets of vertex-indexes of simplex1 and simplex2 
 (sides) where the intersection is an intersection of the corresponding affine 
spaces.

The indexes of simplex2 start with n where n is the lengtrh of simplex1.
So each entry n+j > n in a returnded list represents the j-th element 
of the second simplex: (elt simplex2 j).

An optional tolerance epsilon (which defaults to the global variable *fuzz*) may 
be specified to handle numerical errors in the computation of the intersection."

  (let* ((n                  (length simplex1))
	 (m                  (length simplex2))
	 (n+m                (+ n m))
	 (n+m+1              (1+ n+m))
	 (n+m+2              (1+ n+m+1))
	 (lgr-indexes        (list n+m n+m+1))
	 (kkt-whole-matrix   (kkt-distance-matrix simplex1 simplex2))
	 (kkt-active-matrix  (initially-active n+m))
	 (kkt-rhs            (kkt-distance-right-hand-side n+m))
	 (*fuzz*             (abs epsilon))
	 (+tol               *fuzz*)
	 (-tol               (-  *fuzz*))
	 (lagrangian-1       0d0)
	 (lagrangian-2       0d0)
	 (lagrangian-i       0d0)
	 (kkt-x              kkt-rhs)
	 (search-inactive-sets  ())
	 (new-inactive-sets     ())
	 (discarded-active-sets ())
	 (search-relation       ())
	 (result                ()))

    (loop :for i :below n 
       :do (loop :for j+n :from n :below n+m 
	      :do (push `(,i ,j+n) search-inactive-sets)))
    ;; now all active sets to start with are in search-inactive-sets
    (setf search-inactive-sets (nreverse search-inactive-sets))

    (loop :while (or search-inactive-sets new-inactive-sets)
       :with feasible = T
       :with inactive-set = nil
       :do 
       (if (null search-inactive-sets)
	   (setf search-inactive-sets (nreverse new-inactive-sets)
		 new-inactive-sets    ()))

       (catch 'solve-error 
	 (setf inactive-set     (pop search-inactive-sets)
	       ;; Note: inactive-set is assumed to be sorted.
	       kkt-active-matrix (set-inactive (append inactive-set lgr-indexes)
					       kkt-active-matrix 
					       kkt-whole-matrix)
	       kkt-x             (handler-bind 
				     ((error #'(lambda (condition) 
						 (declare (ignore condition))
						 (format *error-output* 
							 "Solving kkt-matrix ~D failed."
							 kkt-active-matrix)
						 (setf feasible nil)
						 (throw 'solve-error nil))))
				   (laroutines:solve kkt-active-matrix kkt-rhs))
	       lagrangian-1      (aref kkt-x n+m)
	       lagrangian-2      (aref kkt-x n+m+1)
	       feasible          (not (position-if #'(lambda (x) (< x -tol)) ; a fuzzy minusp
						   kkt-x 
						   :end n+m))))

       ;; Note: If an inactive set is unfeasible it may simply be discarded.
       ;;       It is not necessary to remove (activate) any blocking constraint to 
       ;;       make it feasible because the resulting set will show up in 
       ;;       another iteration cycle anyway (See proof below).  
       :when feasible
       :do 
       (if (and (<= -tol lagrangian-1 +tol)(<= -tol lagrangian-2  +tol)) ; fuzzy zerop
	   ;; then the Lagrangians are "zero", hence an intersection was found.
	   (pushnew inactive-set result :test #'set-equals) ;; TODO: subset or superset instead?
	   ;; otherwise there may be an intersection point at some direction.
	   ;; So construct 
	   (let* ((start-idx1  (1+ (car inactive-set)))
		  (start-idx2  (1+ (car (member-if #'(lambda (i) (<= n i)) inactive-set)))))
	     (loop :for i :from start-idx1 :below n+m
		:when (and (not (member i inactive-set)) ; sort of ("active-p" i)
			   (or (< i n) (<= start-idx2 i))) ; i "right-of" first inactive 
		:do ;; i is active and may point to a direction where distance decreases.
		(setf lagrangian-i 0d0)
		(dotimes (j n+m+2)
		  (incf lagrangian-i 
			(* (aref kkt-whole-matrix i j) 
			   (aref kkt-x j))))
		(when (< lagrangian-i  -tol) ; fuzzy minusp
		  ;; then deactivating i decreases distance.
		  (let ((new-active-set (merge 'list (list i) (copy-list inactive-set) #'<)))
		    (unless (member new-active-set 
				    discarded-active-sets 
				    :test #'equal)
		      (push `(,inactive-set ,new-active-set) search-relation)
		      (pushnew new-active-set 
			       new-inactive-sets))))))))
    (values 
     (nreverse result)
     search-relation)))


(defun simplex-intersection-edges (vertices)

;; DONE: Verify correctness!!!
;;       is no more needed, because the function is obsolete.

  (let* ((nothing  0)
	 (edges    ()) ; the (inverted) result
	 (remaining-vertices vertices))

    ;; First find the edges, then the faces, ... etc.
    ;; Each pair with minimal difference makes an edge:
    ;; Problem: Given a vertex... 
    ;;   (1) ...does there always exist another vertex to make an edge?
    ;;   (2) ...how do you handle "multiple" vertices?
    
    ;; Note: each vertex is the sorted dual of its active set.
    ;; assumption: no duplicates in vertices.
    ;; ignored: vertex is sorted. 
    (do* ( (vtx1        (pop remaining-vertices) (pop remaining-vertices))
	   (vtx1-length (length vtx1) (length vtx1))
	   (in-common   nothing nothing)
	   (vtx1-edges  ()  ()) )
	
	 ( (null remaining-vertices) ) ; done

      ;(print `(vtx1 ,vtx1))
      
      (dolist (vtx2-candidate remaining-vertices)

	;(print `(-> vtx2? ,vtx2-candidate))

	(let ( (comm-len (length (intersection vtx1 vtx2-candidate))) )
	  
	  ;(print `(-> common-length ,comm-len))

	  (cond 
	    
	    ((= comm-len vtx1-length)
	     ;; Hey! That's me!
	     ;(print 'found-myself)
	     )

	    ;; This has as much in common as the others found so far:
	    ((= comm-len in-common)
	     ;; collect:
	     ;(print 'collecting)
	     (push vtx2-candidate vtx1-edges))

	    ;; This one has even more in common than the others found so far:
	    ((> comm-len in-common)
	     ;; Start a new collection:
	     ;(print 'collecting-again)
	     (setf in-common  comm-len
		   vtx1-edges (list vtx2-candidate))))))

      ;; Store collected edges:
      (dolist (vtx2 vtx1-edges)
;	(pushnew (sort (union vtx1 vtx2) #'<)  edges
;		 :test #'set-equals
;		 )
	(pushnew (list vtx1 vtx2)  edges
		 :test #'(lambda (a b) (set-equals a b :test #'set-equals)))))

    ;; The result:
    ;; TODO: maybe the intersections of the vertices should be 
    ;; returned instead:
    (nreverse edges)))



(defun simplex-intersection-faces (vertices 
				   &optional 
				   (edges (mapcar #'(lambda (ab) (apply #'union ab))
						  (simplex-intersection-edges vertices))))

;; TODO: Verify correctness!!! (again not needed any more)

  (let* ((nothing  0)
	 (faces    ()) ; the (inverted) result
	 (remaining-edges edges))

    (do* ( (edge           (pop remaining-edges) (pop remaining-edges))
	   (edge-length    (length edge)         (length edge))
	   (in-common      nothing               nothing)
	   (edge-vertices  ()                    ()) )
	
	 ( (null remaining-edges) ) ; done

      (dolist (vtx-candidate vertices)
	
	(unless (subsetp vtx-candidate edge)
	  (let ( (comm-len (length (intersection edge vtx-candidate))) )
	    (cond 
	      
	      ((= comm-len edge-length)
	       ;; Hey! That's me!
	       ;;(print 'found-myself)
	       )

	      ;; This has as much in common as the others found so far:
	      ((= comm-len in-common)
	       ;; collect:
	       ;;(print 'collecting)
	       (push vtx-candidate edge-vertices))

	      ;; This one has even more in common than the others found so far:
	      ((> comm-len in-common)
	       ;; Start a new collection:
	       ;;(print 'collecting-again)
	       (setf in-common  comm-len
		     edge-vertices (list vtx-candidate)))))))

      ;; Store collected faces:
      (dolist (vtx edge-vertices)
	  (pushnew (list vtx edge)  faces
		 :test #'(lambda (a b) (set-equals a b :test #'set-equals))
		 )
	))

    ;; The result:
    ;; TODO: maybe the intersections or unions of the vertices should be 
    ;; returned instead:
    (nreverse faces) )
)




;;; This one links the geometric results intersection-vertices with the topological methods:
(defun simplex-intersection-sides (vertices)

  " Uses combinatorics only to compute the sides (as edges, faces etc.)
 of a simplex-simplex-intersection. 

 The vertices must be specified as the inactive sets at the vertices 
 for some distance-QP. 
 
 The function then returns a sequence seq of sets of inactive sets. 
 Each such element (elt seq i) contains the inactive sets at 
 the i-dimensional affine spaces containing the i-dimensional sides. 

 The topology of the sides is defined by set-inclusion (subsetp). 
 No orientation information is provided."

  (do* ((verts-iter      vertices  vertices)
	(current-sides   ())
	(previous-sides  vertices
			 current-sides)
	(result          (list vertices) 
			 (cons previous-sides result)))

       ((null previous-sides) (nreverse (cdr result)))
    
    (setf current-sides ())

    (do* ((vertex (pop verts-iter) 
		  (pop verts-iter))
	  (vertex-length (length vertex) 
			 (length vertex))
	  (in-common      0 0)   ; Praise thee, FSM!
	  (vertex-sides  ()()))  ; RAMEN
	 
	 ((null verts-iter)) ; this loop is for side effects only.
      
      (dolist (previous-side previous-sides)
	(let ((side-length (length previous-side))
	      (comm-length (length (intersection vertex previous-side))))

	  (cond

	    ((= comm-length side-length))   ; then (subsetp side vertex) => ignore

	    ((= comm-length vertex-length)) ; then (subsetp vertex side) => ignore
      
	    ((= comm-length in-common) ; ya "best candidate" so far
	     (push previous-side vertex-sides)) ; collect
	    
	    ((> comm-length in-common) ; yet a better candidate appeared
	     (setf in-common    comm-length ; start new collection
		   vertex-sides (list previous-side))))))

      (dolist (vertex-side vertex-sides)
	(pushnew (safe-sort (union vertex-side vertex) #'<)
		 current-sides
		:test  #'set-equals)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; From here it would be worthwhile to create a file, say, topology-methods.lisp 

(defun simplex-subset-sign (sorted-subset sorted-superset &key(test #'eql))
  "Returns (-1)^pos where pos is the position of the first element
in sorted-superset which does not occur in sorted-subset at that position or 
0 if both lists are equal.

This will compute the coefficient of sorted-subset as a boundary simplex in 
sorted-superset."

  (simplex-subset-sign-iterative sorted-subset sorted-superset test))


(defun simplex-subset-sign-iterative (sorted-subset sorted-superset test)
  "An iterative implementation of simplex-subset-sign."

  (block function
  (let ((result 1) 
	(subset-iter sorted-subset)
	item-1 )

    (dolist (item-2 sorted-superset)
      (when (null subset-iter)
	(return-from function result))

      (setf item-1 (pop subset-iter))
      (unless (funcall test item-1 item-2)
	(return-from function result))

      (setf result (- result))
      ))
  0
))


(defun simplex-subset-sign-recursive (sorted-subset sorted-superset test result)
  "A tail-recursive implementation of simplex-subset-sign."

  (cond ((null sorted-superset)            0)
	((null sorted-subset)              result)
	((not (funcall test 
		       (car sorted-subset)
		       (car sorted-superset)))  result)
	(t 	 (simplex-subset-sign-aux (cdr sorted-subset)
					  (cdr sorted-superset)
					  test
					  (- result)))))

;; d(1 2 3) = +(2 3) -(1 3) +(1 2)

(defun E-Z-subset-sign (sorted-subset sorted-superset split-item)
  "Computes a boundary coefficient of sorted-subset as a boundary element of 
sorted-superset using the Eilenberg-Zilber-formula. A set 

Each set must be a sorted sequence of integers where exactly one element of 
sorted-superset is missing ion sorted-subset. 

Each specified sequence is split into a subsequence sub-left (or super-left) of all 
elements less than the specified split-item and another sequence sub-rignt (or super-right) 
of the remaining elements.

Then the coefficient of (sub-left sub-right) as a boundary element of (super-left super-right)
is computed using the Eilenberg-Zilber-formula."

  (let ((sub-left    (nreverse (member-if #'(lambda (idx) (< idx split-item)) 
					  (reverse sorted-subset))))
	(sub-right   (member-if #'(lambda (idx) (<= split-item idx)) sorted-subset))
	(super-left  (nreverse (member-if #'(lambda (idx) (< idx split-item)) 
					  (reverse sorted-superset))))
	(super-right (member-if #'(lambda (idx) (<= split-item idx)) sorted-superset)))

    (cond

      ((equal sub-left super-left)
       (if (evenp (length super-left)) ;; then the dimension is odd
	   (- (simplex-subset-sign sub-right super-right)) ; odd dimension
	   (simplex-subset-sign sub-right super-right)))   ; even dimension

      ((equal sub-right super-right)
       (simplex-subset-sign sub-left super-left))

      (T 
       (error "The sets do NOT denote a boundary relation in the prduct of two simplexes.")))))


(defun Lefschetz-subset-sign (sorted-subset sorted-superset size-1)
"It seems that Lefschetz already knew the Eilenberg-Zilber formula before Eilenberg and 
Zilber themselves knew it."
  
  (let ((sub-left    (nreverse (member-if #'(lambda (idx) (< idx size-1)) 
					  (reverse sorted-subset))))
	(sub-right   (member-if #'(lambda (idx) (<= size-1 idx)) sorted-subset))
	(super-left  (nreverse (member-if #'(lambda (idx) (< idx size-1)) 
					  (reverse sorted-superset))))
	(super-right (member-if #'(lambda (idx) (<= size-1 idx)) sorted-superset)))

    (cond

      ((equal sub-left super-left)
       (if (evenp (length super-left)) ;; then the dimension sum is odd
	   (- (simplex-subset-sign sub-right super-right)) ; odd dimension
	   (+ (simplex-subset-sign sub-right super-right))))   ; even dimension

      ((equal sub-right super-right)
       (simplex-subset-sign sub-left super-left))

      (T 
       (error "The sets do NOT denote a boundary relation in the prduct of two simplexes."))
       

      ))
)

(defun transposed-Lefschetz-subset-sign (sorted-subset sorted-superset size-1)
"The boundary coefficient where the left-hand-side and the right-hand-side of the 
tensor are interchanged."
  
  (let ((sub-left    (nreverse (member-if #'(lambda (idx) (< idx size-1)) 
					  (reverse sorted-subset))))
	(sub-right   (member-if #'(lambda (idx) (<= size-1 idx)) sorted-subset))
	(super-left  (nreverse (member-if #'(lambda (idx) (< idx size-1)) 
					  (reverse sorted-superset))))
	(super-right (member-if #'(lambda (idx) (<= size-1 idx)) sorted-superset)))

    (cond

      ((equal sub-left super-left)
       (simplex-subset-sign sub-right super-right))

      ((equal sub-right super-right)
       (if (evenp (length super-right)) ;; then the dimension sum is odd
	   (- (simplex-subset-sign sub-left  super-left )) ; odd dimension (was -)
	   (+ (simplex-subset-sign sub-left  super-left ))))   ; even dimension


      (T 
       (error "The given sets do NOT denote a boundary relation in the prduct of two simplexes."))
      )))


(defun paired-Lefschetz-subset-sign (sorted-subset sorted-superset size-1)
  "A tuple of subset-signs for experiments."

  (list 
   'L  (Lefschetz-subset-sign            sorted-subset sorted-superset size-1)   
   'TL (transposed-Lefschetz-subset-sign sorted-subset sorted-superset size-1)   
   'S (simplex-subset-sign sorted-subset sorted-superset)))


(defun subsets (predicate lst &key (key #'identity))
"Syntax:
  nsubsets predicate lst &key key => list-t, list-nil
Arguments and values:
  predicate -- a one argument predicate
  lst -- a proper list
  list-t, list-nil -- two subssets of lst. 
  key -- a designator for a function with one argument
Descriptiopn:
  Returns two lists, list-t and list-nil 
  one containing all elements satisfying the 
  test, and the other which all elements which do not satisfy 
  the test. 
  The elements are returned in the same order than in the specified list.

  The result may share structure with the given list."

  (values (remove-if (complement predicate) lst :key key)
	  (remove-if predicate lst :key key)))


(defun nsubsets (predicate lst &key (key #'identity))
"Syntax:
  nsubsets predicate lst &key key => list-t, list-nil
Arguments and values:
  predicate -- a one argument predicate
  lst -- a proper list
  list-t, list-nil -- two subssets of lst. 
  key -- a designator for a function with one argument
Descriptiopn:
  Returns two lists, list-t and list-nil 
  one containing all elements satisfying the 
  test, and the other which all elements which do not satisfy 
  the test. 
  The elements are returned in the same order than in the specified list.

  The given list may be destructively modified."

  (let ((removed ()))
    (values (delete-if #'(lambda(item)
			   (if (not (funcall predicate (funcall key item)))
			       ;; note: list must not be modified here (during traversal): 
			       ;; see CLHS 3.6 Traversal Rules and Side Effects
			       (progn (push item removed) T)))
		       lst)
	    (nreverse removed))))



(defun simplex-intersection-boundary (sides)
  "\"Manually\" computes the boundary coefficients of the specified sides, considered manifolds 
such that the boundary represents some orientation of each side.

Sides is a sequence of sets of n-simplexes, each represented by a list of (n+1) integers.

Note that the (transposed) Eilenberg-Zilber-, Lefschetz- or the simple subset-sign 
Give results which do not represent orientation: An edge may have, for example, two 
incoming (outgoing) vertices."

  ;; the following variables contain the corresponding relations 
  ;;   partitioned by dimension.
  ;; In every data structure the lower dimensional cells or cell collections are 
  ;;   left to higher dimensional cells. 
  (with-cons-places ((key . sign)
		     ((vertex-edge-vertex  vertex-edge-edge)   . :sign)
		     ((edge-face-edge      edge-face-face)     . :sign)
		     ((face-volume-face    face-volume-volume) . :sign))
    
    (let* ( ;; the boundary relations from i-cells to i+1-cells
	   ;;   starting with i=0:
	   (relation (mapcar #'(lambda (vertices edges)
				 (let ((result ()))
				   (dolist (vertex vertices)
				     (dolist (edge edges)
				       (when (subsetp vertex edge)
					 (push (list vertex edge) result))))
				   (nreverse result) ))
			     sides
			     (cdr sides)))

	   ;; initialized to the trivial boundary-operator: It's all so zero. It's all so null.
	   (result (mapcar #'(lambda (ith-boundary)
			       (mapcar #'(lambda (pair) `(,pair  . 0))
				       ith-boundary ))
			   relation)))

      (when result

	;; seed the result with vertex-edge-entries to get oriented edges:
	(do* ((visitees    (car result))
	      (vertex-edge (pop visitees) (pop visitees))
	      (other-vertex-edge-test 
	       #'(lambda (this other)
		   (and (not (equal (vertex-edge-vertex this)
				    (vertex-edge-vertex other)))
			(equal (vertex-edge-edge this)
			       (vertex-edge-edge other)))))
	      other-vertex-edge)
	     ((null visitees))

	  (when (zerop (sign vertex-edge))
	    (setf other-vertex-edge 
		  (member vertex-edge visitees
			  :test other-vertex-edge-test))

	    (let* ((ya-vertex-edge (member vertex-edge (cdr other-vertex-edge) 
					   :test other-vertex-edge-test))
		   (involutionary (not ya-vertex-edge)))

	      (when ya-vertex-edge
		(format T 
			"~%Too many vertex edges for ~%~D: ~%  ~D ~%  ~D."
			vertex-edge
			(car other-vertex-edge)
			(car ya-vertex-edge)))
                         
	      (assert involutionary))

	    (when other-vertex-edge
	      (setf other-vertex-edge (car other-vertex-edge))
	      (cond ((zerop (sign other-vertex-edge))
		     (setf (sign other-vertex-edge) +1
			   (sign vertex-edge)       -1))
		    (T
		     (warn "My buddy ~D had a non-zero sign" other-vertex-edge) ; shouldn't happen
		     (setf (sign vertex-edge) (- (sign other-vertex-edge)))))
	      (setf visitees (remove other-vertex-edge visitees :test #'equal)))))
	
	;; now use that seed to compute the rest:
	(mapc #'(lambda (edges-faces faces-volumes)
		  ;; Sets the signs in faces-volumes according to edges-faces.
		  ;; The first "edges-faces" is vertices-edges which have been "seeded" by +/-1 boundaries.

		  (do* ((remaining-fv     faces-volumes      (cdr remaining-fv)) ; iterator 
			(face-volume-seed (car remaining-fv) (car remaining-fv)))

		       ((null remaining-fv) (values))

		    (when (zerop (sign face-volume-seed))
		      ;; correct the sign:
		      (setf (sign face-volume-seed) +1) ; simply set this one to +1

		      ;; now iterate all other faces of this volume:
		      (let* ((volume (face-volume-volume face-volume-seed))

			     ;; collect all faces of that volume not yet visited.
			     (volume-boundary (remove volume remaining-fv
						      :test (complement #'equal)
						      :key  #'face-volume-volume)) ; the entire volume-boundary

			     ;; collect all (half-) edges of these faces
			     (volume-edges    (mapcan #'(lambda (edge-face)
							  (if (member (edge-face-face  edge-face)
								      volume-boundary
								      :key #'face-volume-face
								      :test #'equal)
							      (list edge-face)
							      ()))
						      edges-faces))
			     (remaining-vb    volume-boundary) ; the remaining volume-boundary not yet visited
			     (visitees        (list face-volume-seed)) ) ; will be visited next
			
			;; iterate over faces.
			;;   don't use dolist, because visitees will be modified in the loop body
			(do* ((fv     (car visitees)  (car visitees)) 
			      face
			      face-boundary)

			     ((null visitees)) ; all visits done

			  ;; collect all edges of that face. 
			  (setf visitees      (cdr visitees)
				face          (face-volume-face   fv)
				face-boundary (remove face edges-faces
						      :test (complement #'equal)
						      :key  #'edge-face-face))

			  ;; Iterate all edges and find the corresponding opposite face 
			  ;;    and set opposite sign when approriate.
			  (dolist (edge-face face-boundary)
			    (let* ((other-face-test 
				    #'(lambda (ef oef) 
					(and (equal (edge-face-edge ef) 
						    (edge-face-edge oef))    ; same edge
					     (not (equal (edge-face-face ef) ; different face
							 (edge-face-face oef))))))

				   (other-edge-face ; the face opposite to this edge and boundary to this volume.
				    (member edge-face 
					    volume-edges ; instead of face-boundary
					    :test other-face-test))
				   
				   (other-face-volume ; the opposite face's entry 
				    (find (edge-face-face (car other-edge-face)) ;   to be modified.
					  remaining-vb
					  :test #'equal
					  :key  #'face-volume-face)))
			      
			      (when (member edge-face (cdr other-edge-face) :test other-face-test) ; shouldn't happen.
				(error "edge-face ~D at volume ~D is not involutionary." edge-face volume))
			      
			      (when other-face-volume 
				;; correct the sign when necessary
				(if (zerop (sign other-face-volume))
				    (progn 
				      ;; correct the sign:
				      (setf (sign other-face-volume)
					    (- (/ (* (sign edge-face) 
						     (sign fv)) 
						  (sign (car other-edge-face)))))
				      ;; expand the visited set
				      (push other-face-volume visitees)
				      ;; reduce the volume-boundary (may not be necessary)
				      (setf remaining-vb (remove other-face-volume remaining-vb :test #'equal)))

				    (assert ;; must be correct
				     (zerop (+ (* (sign other-face-volume) (sign (car other-edge-face)))
					       (* (sign edge-face) (sign fv))))))))))))))
	      result
	      (cdr result))
	(ntranspose-boundary result))))) ; OUFFF -- WHAT A HACK!!!



;;; Function to do a boundary-check:

(defun mult-boundary (bd-1 bd-2 &key (test #'eql))
  "Matrix-multiplies the flat boundaries bd-1 and bd-2."

  (with-cons-places (((row-idx col-idx)   . sign))

  (let ((result ())
	(row-indexes (delete-duplicates (mapcar #'row-idx bd-1) :test test))
	(col-indexes (delete-duplicates (mapcar #'col-idx bd-2) :test test)) )

    (dolist (row-index row-indexes)
      (let ((row (members row-index bd-1 :key #'row-idx :test test))
	    col col-entry
	    dot)
	(dolist (col-index col-indexes)
	  (setf col (members col-index bd-2 :key #'col-idx :test test))
	  (when col
	    (setf dot nil)
	    (dolist (row-entry row)
	      (setf col-entry (find (cadar row-entry) col 
				    :test #'(lambda (ridx c)
					      (funcall test ridx (row-idx c)))))
	      (when col-entry
		(unless dot (setf dot 0))
		(incf dot (* (sign row-entry) (sign col-entry)))))
	    (when dot
	      (push (cons `(,row-index ,col-index) dot)
		    result))))))
    (nreverse result))))


(defun flat-boundaryp (bd &key (test #'eql))
"Tests if bd is a boundary by computing the matrix square of bd.
bd must be partitioned"
  (with-cons-places ((:key . value))
    (let* ((flat-bd (reduce (rcurry #'chain+ :test test)  bd))
	   (flat-bd-sq (mult-boundary flat-bd flat-bd :test test)))
      (null (member 0 flat-bd-sq
		    :key   #'value
		    :test  (complement #'equalp))))))

(defun boundaryp (bd &key (test #'eql)  &allow-other-keys)
  "Tests if bd is a boundary by computing the matrix squares m-i+1 x m-i of each two 
successive matrices m-i m-i+1 in bd.

Note the inverse order: if (A B) is a subsequence in bd then B x A will be computed."

  (every #'(lambda (bd-i+1 bd-i)
	      (every (compose #'zerop #'matrix-entry-value)
		     (mult-boundary bd-i+1 bd-i :test test)))
	 (cdr bd)
	 bd ))

(defun co-boundaryp (bd &key (test #'eql) &allow-other-keys)
  "Tests if bd is a boundary by computing the matrix squares m-i+1 x m-i of each two 
successive matrices m-i+1 m-i in bd.

Note the maintenance of order: if (A B) is a subsequence in bd then A x B will be computed."
  (every #'(lambda (bd-i+1 bd-i)
	      (every (compose #'zerop #'matrix-entry-value)
		     (mult-boundary bd-i+1 bd-i :test test)))
	 bd
	 (cdr bd)))


(defun transpose-boundary (bd)
  (with-cons-places (((row-idx col-idx)   . sign))
    (mapcar #'(lambda (dim-bd)
		(mapcar #'(lambda (entry)
			    (let ((result (copy-tree entry)))
			      (rotatef (row-idx result) (col-idx result))
			      result))
			dim-bd))
	    bd)))

(defun ntranspose-boundary (bd)
  (with-cons-places (((row-idx col-idx)   . sign))
    (mapc #'(lambda (dim-bd)
	      (mapc #'(lambda (entry)
			(rotatef (row-idx entry) (col-idx entry)))
		    dim-bd))
    bd))
  bd)


     
;;; Again that Simple Test-Case:
;;; - Two Triangles in IR^2
;;;
;;;                   * 5:(0,3)
;;;                  / \
;;;                 /   \
;;;                / 2345\
;;;               /   * 2:(0,2)
;;;              /   / \   \
;;;             /   /   \   \
;;;            /   / 234 \   \
;;;  3:(-1,1) *---x-------x---* 4:(1,1)
;;;              /0234     \1234
;;;             /           \
;;;            /             \
;;;  0:(-1,0) *---------------* 1:(1,0)
;;;
(let ((vertices 
       (simplex-intersection-vertices
	#(#(-1d0 0d0) #(1d0 0d0) #(0d0 2d0))
	#(#(-1d0 1d0) #(1d0 1d0) #(0d0 3d0))
	1d-13)))
  (print `(vertices ,vertices))
  (print `(sides    ,(simplex-intersection-sides vertices)))
)

;;; A Simple Test-Case:
;;; - Two Triangles in IR^2 one is flat, one is tall.
;;;
;;;                         * 2:(0,3)
;;;                       /| |\
;;;                       /   \
;;;                      /  *  5:(0,2) 0125
;;;                     / /| |\ \
;;;                    //[-] [+]\\
;;;              0235 x           x 1245
;;;                 //             \\  
;;;               / /(-)         (+)\ \ 
;;;             /  /       [+]       \  \
;;;  3:(-1,1) *---x-------------------x-->* 4:(1,1)
;;;              /0234                 \1234
;;;             /                       \
;;;            /            (+)          \
;;;  0:(-1,0) *-------------------------->* 1:(1,0)
;;;
;;; The sides found:
;;;
;;;        02|35                012|5              12|45
;;;            *------012|35------*-----012|45-----* 
;;;            |      (1d)              (1d)       |
;;;            |                                   |
;;;            |                                   |
;;;       02|345              012|345              12|345
;;;        (1d)|               (2d)                | (1d)
;;;            |                                   |
;;;            |                                   |
;;;            *---------------012|34--------------*
;;;        02|34                (1d)              12|34




(let* ((vertices 
	(simplex-intersection-vertices
	 #(#(-1d0 0d0) #(1d0 0d0) #(0d0 3d0))
	 #(#(-1d0 1d0) #(1d0 1d0) #(0d0 2d0))
	 1d-10)) ; try 1d-13
       (sides        (simplex-intersection-sides vertices))
       (boundary     (simplex-intersection-boundary sides)))
  (print `(vertices ,vertices))
  (print `(sides    ,sides))
  (print `(bd       ,boundary))
  (print `(bd^2     ,(mapcar #'(lambda (a b) (mult-boundary a b :test #'equal))
			     (cdr boundary)
			     boundary)))
  (print `(is boundary? ,(if (boundaryp boundary)
			    'YES!
			    'NO!)))
)


;;; now a 3D test case:

;;; Again that Simple Test-Case:
;;; - Two Triangles in IR^2
;;;
;;;                   * 5:(0,3)
;;;                  / \
;;;                 /   \
;;;                / 2345\
;;;               /   * 2:(0,2)
;;;              /   / \   \
;;;             /   /   \   \
;;;            /   / 234 \   \
;;;  3:(-1,1) *---x-------x---* 4:(1,1)
;;;              /0234     \1234
;;;             /           \
;;;            /             \
;;;  0:(-1,0) *---------------* 1:(1,0)
;;;


;;; Again that Simple Test-Case:
;;; - Two 3-simplexes in IR^3
;;; - no critical vertices
;;;
;;;                     * 6:(0|3|1)
;;;                    /|\
;;;                   / * \7:(0|2.5|4)
;;;                  / / \ \
;;;                 //  * 2:(0|2|0)
;;;                /   /|\   \
;;;               /   / x \   \
;;;              /   / /|\ \   \
;;;  4:(-1|1|1) *-----x-|-x-----* 5:(1|1|1)
;;;                /    |3:(0|0.5|3)
;;;               /   . * .   \
;;;              /.          . \
;;;  0:(-1|0|0) *-------+-------* 1:(1|0|0)
;;;                   (000)

(let* ((vertices 
	(simplex-intersection-vertices
	#(#(-1d0 0d0 0d0) #(1d0 0d0 0d0) #(0d0 2d0 0d0) #(0d0 0.5d0 3d0))
	#(#(-1d0 1d0 1d0) #(1d0 1d0 1d0) #(0d0 3d0 1d0) #(0d0 2.5d0 4d0))
	1d-13
	 ))
       (sides    (simplex-intersection-sides vertices))
       (boundary (simplex-intersection-boundary sides))
       )
  (print `(vertices ,vertices))
  (print `(sides    ,sides))
  (print `(bd       ,boundary))
  (print `(bd^2     ,(mapcar #'(lambda (a b) (mult-boundary a b :test #'equal))
			     (cdr boundary)
			     boundary)))
  (print `(is boundary? ,(if (boundaryp boundary)
			    'YES!
			    'NO!)))
)



;;; The upper test case with critical vertices (all "bottom" vertices with 
;;; third coordinate zero):
;;; - Two 3-simplexes in IR^3
;;;
;;;                     * 6:(0|3|0)
;;;                    /|\
;;;                   / * 7:(0|2.5|4)
;;;                  / / \ \
;;;                 //  * 2:(0|2|0)
;;;                /´  /|\  `\
;;;               /   / x \   \
;;;              /   / /|\ \   \
;;;  4:(-1|1|0) *-----x-|-x-----* 5:(1|1|0)
;;;                /    |    \
;;;               /   . * 3:(0|0.5|3)
;;;              /. '       '  . \
;;;  0:(-1|0|0) *-------+-------* 1:(1|0|0)
;;;                   (000)

(catch 'test-failed 
  (handler-bind ((simple-error #'(lambda (err)
				   (terpri *error-output*)
				   (princ "Test failed:" *error-output*)
				   (terpri *error-output*)
				   (apply #'format *error-output*
					  (simple-condition-format-control err)
					  (simple-condition-format-arguments err))
				   (throw 'test-failed err))))

    (let* ((vertices 
	    (simplex-intersection-vertices
	     #(#(-1d0 0d0 0d0) #(1d0 0d0 0d0) #(0d0 2d0 0d0) #(0d0 0.5d0 3d0))
	     #(#(-1d0 1d0 0d0) #(1d0 1d0 0d0) #(0d0 3d0 0d0) #(0d0 2.5d0 4d0))
	     1d-13
	     ))
	   (dummy-1   (print `(vertices ,vertices)))
	   (sides     (simplex-intersection-sides vertices))
	   (dummy-2   (print `(sides    ,sides)))
	   (boundary  (simplex-intersection-boundary sides))
	   (dummy-3   (print `(bd       ,boundary))))

      (declare (ignore dummy-1 dummy-2 dummy-3))

      (print `(bd^2     ,(mapcar #'(lambda (a b) (mult-boundary a b :test #'equal))
				 (cdr boundary)
				 boundary)))
      (print `(is boundary? ,(if (boundaryp boundary)
				 'YES!
				 'NO!)))
      (values))))


;; fails with 
'(VERTICES
  (0245      1245     2456     02457    02345    12345   12457    23456    23457   24567))
'(SIDES
  ((0245     1245     2456     02457    02345    12345   12457    23456    23457   24567)) ; 0d
  ((234567   12456    123456   123457   124567   01245   02456    012345   012457  023456   023457   024567)) ;1d
  ((124567   123456   012345   023456   012457   024567  1234567  0234567  012456  0123456  0123457  0124567)) ; 2d
  ((1234567  0123457  0234567  0124567  0123456  01234567)) ; 3d
  ((01234567)))                                             ; 4d ???
;; next step was remove-duplicates...:
;; still fails with 
'(VERTICES
 ((0 2 4 5) (1 2 4 5) (2 4 5 6) ; purged away
  ;; the following are kept:
  (0 2 4 5 7) (0 2 3 4 5) (1 2 3 4 5)
  (1 2 4 5 7) (2 3 4 5 6) (2 3 4 5 7) (2 4 5 6 7))) 
'(PURGED-VERTICES
 ((0 2 4 5 7) (0 2 3 4 5) (1 2 3 4 5) (1 2 4 5 7) (2 3 4 5 6) (2 3 4 5 7)
  (2 4 5 6 7))) 
'(SIDES
 (((0 2 4 5 7) (0 2 3 4 5) (1 2 3 4 5) (1 2 4 5 7) (2 3 4 5 6) (2 3 4 5 7)
   (2 4 5 6 7))
  ((2 3 4 5 6 7) (1 2 4 5 6 7) (1 2 3 4 5 6) (1 2 3 4 5 7) (0 1 2 3 4 5)
   (0 2 3 4 5 6) (0 1 2 4 5 7) (0 2 3 4 5 7) (0 2 4 5 6 7))
  ((1 2 3 4 5 6 7) (0 1 2 3 4 5 6) (0 1 2 4 5 6 7) (0 1 2 3 4 5 7)
   (0 2 3 4 5 6 7))
  ((0 1 2 3 4 5 6 7)))) 
;; Problem: there are duplicate vertices and edges at the common bottom face:
;; when the bottom face of one simplex is moved slightly up- or downwards the 
;; duplicate vertices disappear and boundary computation doesn't fail.

;; Idea: find connected components and take the larges.
(defun find-components (relation &key (test #'eql) (relating #'car) (related #'cdr))

  (do ((result ())
       (remaining (mapcar #'(lambda (pair)
			      (list (funcall relating pair)
				    (funcall related pair)))
			  relation)))

      ((null remaining) (nreverse result))

    (do* ((current (pop remaining))
	  (new     current))

	 ((null new)
	  (push (delete-duplicates current
				   :test test)
		result))

      (multiple-value-bind (newer visitees)
	  (subsets #'(lambda (pair)
		       (or (member (car  pair) new :test test)
			   (member (cadr pair) new :test test)))
		   remaining)

	(setf newer      (remove-duplicates 
			  (set-difference (reduce #'append newer)
					  current
					  :test test)
			  :test test)
	      new        newer
	      current    (append current newer)
	      remaining  visitees)))))


(defparameter sides (reduce #'append
 '(((0 2 4 5 7) (0 2 3 4 5) (1 2 3 4 5) (1 2 4 5 7) (2 3 4 5 6) (2 3 4 5 7)
   (2 4 5 6 7))
  ((2 3 4 5 6 7) (1 2 4 5 6 7) (1 2 3 4 5 6) (1 2 3 4 5 7) (0 1 2 3 4 5)
   (0 2 3 4 5 6) (0 1 2 4 5 7) (0 2 3 4 5 7) (0 2 4 5 6 7))
  ((1 2 3 4 5 6 7) (0 1 2 3 4 5 6) (0 1 2 4 5 6 7) (0 1 2 3 4 5 7)
   (0 2 3 4 5 6 7))
  ((0 1 2 3 4 5 6 7))))) 

(defparameter relation (mapcan #'(lambda (a)
				   (mapcan #'(lambda (b)
					       (if (subsetp a b :test #'equalp)
						   (list (cons a b))
						   () ))
					   sides ))
			       sides))
				   

;;; a small test case:
(print (find-components '((1 . 2) (2 . 3) (4 . 5)  (5 . 6) (7 . 8) (1 . 8))))

;;; This find-component ides, however, fails:
(print relation)
(print (find-components relation))
(print (length (find-components relation)))
;; the length is 1 as the topological space generated by that relation is connected.


;;; Next idea:
;;; Find /minimal/ inactive sets representing vertices:
;;; Conjecture: Critical vertices show up as one miniml inactive set and a family of 
;;;             its supersets. 
;;;
;;; Question 1: Does the modified active set method generate all possible 
;;;             critical vertices (provided no numerical errors occur)?
;;;
;;; Question 2: How do numerical errors influence the answer to question 1?
;;;
;;; Question 3: Do edges-vertices-... always show up in the computed sides?
;;;             (In other words: does every "next side" starting with a set 
;;;             of sides and a set of vertices increase dimension by one.)
;;; Question 4: Does the Lefschetz boundary work with results having critical 
;;;             vertices?


(defun proper-subsetp (list-1 list-2 &key key test test-not)
  "Returns true, if list-1 is a subset of list-2 but list-2 is not a subset of list-1.
otherwise false."

  (let ((test (cond (test       test)
		    (test-not   (complement test-not))
		    (t          #'eql)))) ; 17.2.1 in HyperSpec.
    (and (not (eq list-1 list-2)) ; speed up test: costs constant time.
	 (subsetp list-1 list-2 :key key :test test)
	 (not (subsetp list-2 list-1 :key key :test test)))))


(defun minimal-sets (sets &key (test #'eql))
  "Syntax:
  minimal-sets list &key test ==> result-list
Arguments and values:
  list -- a proper list of proper lists
  test -- a designator for a function of two arguments that returns a generalized boolean.
  result-list -- a proper list of lists.

Description:
  Returns a list of all elements of the specified list which have no 
  proper subset within that list. 
  The order of elements in list is maintained.

Example:
  (minimal-sets '((1 2 3) (1 2 3 4) (2 3 4 5)))
  ==> ((1 2 3) (2 3 4 5))

  (minimal-sets '((1 2 3) (3 2 1) (2 3 4 5)))
  ==> ((1 2 3) (3 2 1) (2 3 4 5))

  (minimal-sets '((1 2 3) (1 2 3 4) (2 3 4 5) nil ...))
  ==> (nil)"

  (remove-if #'(lambda (set)
		 (some #'(lambda (better-set) 
			   (proper-subsetp better-set set :test test))
		       sets))
	     sets))

(print `(purged ,(minimal-sets '((1 2 3) (2 3 4) (1 2) (2 3)))))
(print `(purged ,(minimal-sets '((1 2 3) (2 3 4) () (1 2) (2 3)))))
(print (minimal-sets '((1 2) (1 2 3) (3 2 1) (3 1) (2 3 4 5))))

;;; With purges it seems to work:
(let* ((vertices 
	(simplex-intersection-vertices
	 #(#(-1d0 0d0 0d0) #(1d0 0d0 0d0) #(0d0 2d0 0d0) #(0d0 0.5d0 3d0))
	 #(#(-1d0 1d0 0d0) #(1d0 1d0 0d0) #(0d0 3d0 0d0) #(0d0 2.5d0 4d0))
	 1d-13
	 ))
       (dummy-1   (print `(vertices ,vertices)))
       (sides     (simplex-intersection-sides (minimal-sets vertices)))
       (dummy-2   (print `(sides    ,sides)))
       (boundary  (simplex-intersection-boundary sides))
       (dummy-3   (print `(bd       ,boundary))))

  (declare (ignore dummy-1 dummy-2 dummy-3))

  (print `(bd^2     ,(mapcar #'(lambda (a b) (mult-boundary a b :test #'equal))
			     (cdr boundary)
			     boundary)))
  (print `(is boundary? ,(if (boundaryp boundary)
			     'YES!
			     'NO!)))
  (values)
  )


;;; three test cases to find the search paths of the active set method:
;;; (simplex-intersection vertices was modified to return the search-paths 
;;; as a secondary value)
;;;
;;;                            -   -  - - - ---------* 3:(3,1.1) 3-4 slanted up
;;;         4:(-1,1) *-------------------------------* 3:(3,1)   3-4 horizontal
;;;                    \                           / * 3:(3,0.9) 3-4 slanted down
;;;                      \                       //
;;;                        \                   //
;;; *------------------------* 1:(0,0)       /
;;; 0:(-3,0)                   \           /
;;;                              \       /
;;;                                \   / 
;;;                                  * 2:(1,-1)
;;;                                  
;;; this shows, that not all inactive sets are computed.

(let (vertices relation)

  (multiple-value-setq (vertices relation)
    (simplex-intersection-vertices
     #|the 0-1 line:       |#  #(#(-3.0d0 0.0d0) #(0.0d0 0.0d0))
     #|the 2-3-4 triangle: |#  #(#(1.0d0 -1.0d0) #(3.0d0 1.0d0  #|ONE !|#) #(-1.0d0 1.0d0))))
  (terpri)(princ '3-4-horizontal\:)
  (print 'vertices)(princ '\:)
  (mapcar #'print vertices)
  (print 'relation)(princ '\:)
  (mapcar #'print relation)

  (multiple-value-setq (vertices relation)
    (simplex-intersection-vertices
     #|the 0-1 line:       |#  #(#(-3.0d0 0.0d0) #(0.0d0 0.0d0))
     #|the 2-3-4 triangle: |#  #(#(1.0d0 -1.0d0) #(3.0d0 1.1d0 #|ONE.ONE !|#) #(-1.0d0 1.0d0))))
  (terpri)(princ '3-4-slanted\ up\:)
  (print 'vertices)(princ '\:)
  (mapcar #'print vertices)
  (print 'relation)(princ '\:)
  (mapcar #'print relation)

  (multiple-value-setq (vertices relation)
    (simplex-intersection-vertices
     #|the 0-1 line:       |#  #(#(-3.0d0 0.0d0) #(0.0d0 0.0d0))
     #|the 2-3-4 triangle: |#  #(#(1.0d0 -1.0d0) #(3.0d0 0.9d0 #|ZERO.NINE !|#) #(-1.0d0 1.0d0))))
  (terpri)(princ '3-4-slanted\ down\:)
  (print 'vertices)(princ '\:)
  (mapcar #'print vertices)
  (print 'relation)(princ '\:)
  (mapcar #'print relation)
)
     
;;; Proof that the minimal inactive set always shows up:

#||

Let {a_1,...,a_n} and {b_1,...,b_m} be two vertex sets of simplexes.
Let {a_i1,...,a_ik} and {b_j1,...,b_jr} define an intersection 
point p. 
Let {a'_i'1,...,a'_i'k'} and {b'_j'1,...,b'_j'r'} define the same 
intersection point p. 
Then p is in the convex closure of each of the four specified sets.
By A and B being affine independent the convex closures of the intersections 
  cl({a_i1,...,a_ik} \cap {a'_i'1,...,a'_i'k'})
  cl({b_j1,...,b_jr} \cap {b'_j'1,...,b'_j'r'})
are the intersection of the convex closures:
  cl({a_i1,...,a_ik}) \cap cl({a'_i'1,...,a'_i'k'}).
  cl({b_j1,...,b_jr}) \cap cl({b'_j'1,...,b'_j'r'}).

By p being an intersection point all these closures containe p.
Hence the intersection of all (inactive) vertex sets defining 
the point p also defines the point p and is therefor minimal.
Note that the excessive vertices in the sets are active.

Now we know that such minimal set exists and is unique.
We now have to show, that the algorithm computes this set, too.
 (Assuming no numerical errors occur)

Stmt. The algorithm iterates over all feasible subsets of the minimal 
      optimal inactive sets.

Proof: The sets {a,b} s.t. a \in A and b\in B are the minimal feasible 
sets and are all generated at the initialization. 

Let S={i_1,...,i_k, j_1,..., j_r} be a feasible subset of of the vertex indices 
of vertices of A and B. Let S contain no active index. 
We will later see that we can assume without loss of generaltiy, that S is a 
minimal inactive set of the intersection point of the two simplexes 
A_S={a_i_1, ... , A_j_k} and B_S={b_j_1, ... , b_j_r}. 

Then by removing one vertex (index) t from S 
the remaining set may or may not be feasible. 

Now there always exists a vertex v in S which can be removed such 
that the remaining set S - {v} is feasible and contains no active set.

Proof:
See doc/active-set-proofs.tex

But then the distance function of that set S-{v} is bigger than S, hence 
the gradient is negative, or, in other words, the Lagrangian of the activation 
is positive. 

But then v will be added to S-{v} if S-{v} is present in the search set. 

||#

;;; ----------------- End of Simplexes ---------------------------------------


;;; ------------------Start of polytopes -------------------------------------
;;; Now we want to handle polytopes:
;;; We triangulate a polytope complex into simplexes and compute a 
;;; simplex-by-simplex-intersection.


;;; Conventions: 
;;; - complexes are passed with two arguments:          boundary        cells.
;;; - result complexes are returned as multiple values: result-boundary result-cells.
;;; - Boundaries and cells always start with the lowest-dimensional elements.
;;; - For each function (fx boundaries cells) there should be a function (co-fx co-boundary co-cells)
;;;   with dimension reversed input and accordingly reversed output. 
;;;   co-boundary is not the mathematical co-boundary (reversed order /and/ transposed) but only
;;;   the sequence of boundaries in reversed order. 
;;  - Prefixing with co- indicates reversed order.

;;; Combinatorial triangulation of a complex: 
;;;   1: Take the boundary dA of a cell A.
;;;   2: Take one vertex a from A
;;;   3: replace A by the "cone" <a> * dA.
;;;       where  * stands for tuple-wise multiplication : <a> * <x y ... z> = <a x y ... z>
;;; It is easy to prove that the (algebraic) boundary of that cone 
;;; is equal to the boundary of A (dA =  d(<a> * dA) if dA is a simplicial complex:
;;;   Let dA be a simplicial complex without boundary (ddA=0).
;;;   Compute d(<a> * dA) where i ranges from i to n, the number of boundary cells:
;;;   d(<a> * dA) = d (<a> * sum i : (s_i <v_i0 ... v_id>)       s_i is the sign of <v_i0 ... v_id>
;;;               = d ( sum i : s_i <a>*<v_i0 ... v_id> )        "*" distributes
;;;               = d ( sum i : s_i <a v_i0 ... v_id> )          
;;;               = sum i : s_i d<a v_i0 ... v_id>               d is linear
;;;               = sum i : s_i (<v_i0 ... v_id>                 here a has been stripped off ...
;;;                              - <a v_i1 ... v_id>                ... and elsewere
;;;                              + <a v_i0 v_i2 ... v_id> - ...)    ... a is kept.
;;;               = sum i : s_i <v_i0 ... v_id>  - sum i : s_i (<a v_i1 ... v_id> - <a v_i0 v_i2 ... v_id> + ...)
;;;               = sum i : s_i <v_i0 ... v_id>  - sum i : s_i (<a v_i1 ... v_id> - <a v_i0 v_i2 ... v_id> + ...)
;;;               = dA                           - sum i : <a> * s_i (<v_i1 ... v_id> - <v_i0 v_i2 ... v_id> + ...)
;;;               = dA                           - <a> * sum i : s_i d(<v_i0 ... v_id>)
;;;               = dA                           - <a> * d(sum i : s_i <v_i0 ... v_id>)
;;;               = dA                           - <a> * ddA
;;;               = dA                           - 0
;;;               = dA
;;; On the other hand that cone is simplicial if dA is simplicial.
;;;
;;; Shorter version where dA need not be simplicial:
;;;   d(<a> * dA) = d<a> * dA  +/-  <a> * ddA
;;;               =  <>  * dA  +/-  <a> * 0
;;;               = dA
;;;
;;;   4: It is convenient to set the boundary of every cone's element aE in cA to 
;;;         d(aE) := d(a)*E - a*d(E) = E - a*d(E)
;;;       because of 
;;;         dd(aE) = d(E) - d(a*d(E))
;;;                = d(E) - (d(a)*d(E)  +/-  a*dd(E))
;;;                = d(E) - <>*d(E)  +/- 0
;;;                = d(E) - d(E) 
;;;                = 0.
;;;   5: Update the boundary AND THE CO-BOUNDARY of A.
;;;   

(defun choose (x) (first x))
;; alternative: (defun choose (x) (when x (elt (random (length x)) x)))

(defun printc (object &optional output-stream)
  "An \"aesthetic\" print."
  (let ((stream ; this works around an sbcl-bug:
	 (cond ((null output-stream) *standard-output*)
	       ((eq output-stream t) *terminal-io*)
	       (t output-stream))))
    (terpri stream)
    (write object
	   :stream   stream
	   :readably nil
	   :escape   nil )
    (write-char #\space stream))
  object)




;;; And now some test-cases:
;;; A cube:
;;;
;;;          ,.g -------------->,.h
;;;        ´/' ^              ´/' ^  
;;;       /    :  T          /    |  
;;;     /      :           /      |   
;;;   e ---------------> f        |   
;;;   ^        :     (H) ^        |   
;;;   |  (L)   :         |   R    | 
;;;   |        : F       |        | 
;;;   |      ,.c  - - - -|- - ->,.d
;;;   |    ´.'           |    ´/'
;;;   |   .       (B)    |   /
;;;   | .                | /
;;;   a ---------------> b 
;;; 
;;; edges: ab ac ae
;;;        bf bd
;;;        cd cg
;;;        ef eg
;;;        dh
;;;        fh
;;;        gh
;;;
;;; faces: F (ront), H (ind), T (op), B (ottom), L (eft), R (ight)
;;;        All faces have counterclockwise orientation on the projection.
;;; volume: V
;;;
;;; Boundary:
;;;
;;; d(V) = T - B  +  F - H  +  R - L
;;;
;;; d(T) = ef + fh - gh - eg
;;; d(B) = ab + bd - cd - ac
;;; d(F) = ab + bf - ef - ae
;;; d(H) = cd + dh - gh - cg
;;; d(R) = bd + dh - fh - bf
;;; d(L) = ac + cg - eg - ae
;;;
;;; d(ab) = b - a
;;; ...
;;; d(gh) = h - g.

(defparameter *sample-cube-cells*        
  '((\a \b \c \d \e \f \g \h)  ; 0-cells
    (|ab| |ac| |ae|   
     |bf| |bd|   
     |cd| |cg|   
     |ef| |eg|   
     |dh| |fh| |gh|)         ; 1-cells
    (T  B  F  H  R  L )      ; 2-cells
    (V))               ; 3-cell
)

(defparameter *sample-cube-boundary* 
  '((;; d:X1 -> X0
     ((|ab| \a) . -1) ((|ab| \b) . +1) ; d(ab) = b - a
     ((|ac| \a) . -1) ((|ac| \c) . +1) ; d(ac) = c - a
     ((|ae| \a) . -1) ((|ae| \e) . +1) ; ...
     ((|bf| \b) . -1) ((|bf| \f) . +1)
     ((|bd| \b) . -1) ((|bd| \d) . +1)
     ((|cd| \c) . -1) ((|cd| \d) . +1)
     ((|cg| \c) . -1) ((|cg| \g) . +1)
     ((|ef| \e) . -1) ((|ef| \f) . +1)
     ((|eg| \e) . -1) ((|eg| \g) . +1)
     ((|dh| \d) . -1) ((|dh| \h) . +1)
     ((|fh| \f) . -1) ((|fh| \h) . +1)
     ((|gh| \g) . -1) ((|gh| \h) . +1))
    
    (;; d:X2 -> X1
     ((T  |ef|) . +1) ((T  |fh|) . +1) ((T  |gh|) . -1) ((T  |eg|) . -1) ; d(T) = ef + fh - gh - eg
     ((B  |ab|) . +1) ((B  |bd|) . +1) ((B  |cd|) . -1) ((B  |ac|) . -1) ; d(B) = ab + bd - cd - ac
     ((F  |ab|) . +1) ((F  |bf|) . +1) ((F  |ef|) . -1) ((F  |ae|) . -1) ; d(F) = ab + bf - ef - ae
     ((H  |cd|) . +1) ((H  |dh|) . +1) ((H  |gh|) . -1) ((H  |cg|) . -1) ; d(H) = cd + dh - gh - cg
     ((R  |bd|) . +1) ((R  |dh|) . +1) ((R  |fh|) . -1) ((R  |bf|) . -1) ; d(R) = bd + dh - fh - bf
     ((L  |ac|) . +1) ((L  |cg|) . +1) ((L  |eg|) . -1) ((L  |ae|) . -1) ; d(L) = ac + cg - eg - ae
     )
    (;; d:X3 -> X2
     ((V T) . +1) ((V B) . -1) ((V F) . +1) ((V H) . -1) ((V R) . +1) ((V L) . -1) ; d(V) = T - B  +  F - H  +  R - L
     ))
  )

(let* ((cells    *sample-cube-cells*)
       (boundary *sample-cube-boundary*))

  (printc `(bd^2     ,(mapcar #'(lambda (a b) (mult-boundary a b :test #'equal))
			     (cdr boundary)
			     boundary)))

  (printc `(is boundary? ,(if (boundaryp boundary)
			     'YES!
			     'NO!)))

  (catch 'failed
    (handler-case 
	(printc `(triangulated ,(complex-triangulation cells boundary #|:test #'equal|#)))
      (error (condition)
	(format *error-output* "~%An error has occured:~%~A" condition)
	(throw 'failed 'failed)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Triangulation:

;;; Some tests needed to dig into tensor products:

(defun tree-test (test a b)
  (labels ((tree-test (a b)
	     (or (eq a b)
		 (ignore-errors (funcall test a b)) ; what an ugly hack;
		 (and (not (atom a)) (not (atom b))
		      (tree-test (car a) (car b))
		      (tree-test (cdr a) (cdr b))))))
    (tree-test a b)))

(defun list-test (test list-1 list-2)
  (or (eq list-1 list-2)
      (when (= (length list-1) (length list-2))
	(every test list-1 list-2))))


(defun tensor-test (test list-1 list-2)
  (or (eq list-1 list-2)
      (ignore-errors (funcall test list-1 list-2)) ;what an ugly hack!!
      (and (listp list-1)
	   (listp list-2)
	   (= (length list-1) (length list-2))
	   (every test list-1 list-2)
	   T)))


(defun simplex-test (test list-1 list-2)
  "simplex-test test list-1 list-2 -> generalized-boolelan

Successively applies the test to each element in list-1 
in order to the first element of list-2 that satisfies the 
test, if such element exists. In this case that element will 
be excluded from subsequent tests. 
If no such element exists or if not all elements of list-2 
have been exhausted simplex-test returns nil. "

  (labels 
      ((test-aux (list-1 list-2)
	 (let (top pos)
	   (cond ( (eq list-1 list-2) 
		  T)
		 ( (or (atom list-1) (atom list-2))
		  nil )
		 ( (funcall test (setf top (car list-1)) (car list-2))
		  (test-aux (cdr list-1) (cdr list-2)) )
		 ( (setf pos (position top list-2))
		  (test-aux (cdr list-1) (remove-nth pos list-2)))))))
    (test-aux list-1 list-2)))


(defun permutation-sign-recursive (list-1 list-2 &key (test #'eql))
  (permutation-sign-recursive-aux list-1 list-2 1 test))

(defun permutation-sign-recursive-aux (list-1 list-2 result test &aux pos)

  "The recursive version."
  (cond ((null list-1) result)

	((funcall test (car list-1) (car list-2))
	 (permutation-sign-recursive-aux (cdr list-1) (cdr list-2) result test))

	((setf pos (position (car list-1) list-2 :test test))
	 (permutation-sign-recursive-aux (cdr list-1)
					 (remove-nth pos list-2)
					 (if (evenp pos) 
					     result
					     (- result))
					 test))))


(defun permutation-sign (list-1 list-2 &key (test #'eql))

  "Returns the sign of the permutation from list-1 to list-2.

It is an error to provide two lists s.t. 
    (subsetp list-1 list-2 :test test)
and (subsetp list-2 list-1 :test test)
don't hold or if either list contains duplicates."

  (permutation-sign-aux list-1 list-2  1  test))

(defun permutation-sign-aux (list-1 list-2 result test)
    (cond ((eq list-1 list-2) result)
	  ((funcall test (car list-1) (car list-2))
	   (permutation-sign-aux (cdr list-1) (cdr list-2) result test))
	  (t
	   (let ((pos (position (car list-1) list-2 :test test)))
	     (npermutation-sign-aux (cdr list-1) 
				    (remove-nth pos list-2)
				    pos
				    (if (oddp pos) (- result) result)
				    test)))))


(defun npermutation-sign-aux (list-1 list-2 maxpos result test)
  "Returns the sign of the permutation from list-1 to list-2, 
thereby destroying the cdr chain of list-2 up to maxpos.

It is an error to provide two lists s.t. 
    (subsetp list-1 list-2 :test test)
and (subsetp list-2 list-1 :test test)
don't hold or if either list contains duplicates."
  (let ((pos nil))
    (cond ((eq list-1 list-2) result)
	  ((funcall test (car list-1) (car list-2))
	   (npermutation-sign-aux (cdr list-1) (cdr list-2) (1- maxpos) result test))
	  ((null (setf pos (position (car list-1) list-2 :test test)))
	   (error "list-1 is not a subset of list-2."))
	  ((< pos maxpos) ; it is safe to delete 
	   (npermutation-sign-aux (cdr list-1) 
				  (delete-nth pos list-2)
				  (1- maxpos)
				  (if (oddp pos) (- result) result)
				  test))
	  (t ; cannot be deleted, must be removed.
	   (npermutation-sign-aux (cdr list-1) 
				  (remove-nth pos list-2)
				  pos
				  (if (oddp pos) (- result) result)
				  test)))))


#|Correctness of permutation-sign:
 Let list-1 be an arbitrary list.
   Then with (permutation-sign list-1 list-1) the algorithm compares 
   each element elt-i in list-1 with itself, satisfying the test.
   Then result will never change its initial value of 1.
   So the result for the identity permutation is correctly 1.
Let list-2 be a permutation of list-1.
  Let a1 be the first element in list-1.
  If a1 is the first element of list-2, too 
    then the permutation sign of list-1 list-2 is the same as the 
    permutation sign of (rest list-1) and (rest list-2). 
    This is correctly handled by the loop expression. 
  If a1 is /not/ the first element in list-2 it can be moved  
    to the first place by pos = (position elt-1 list-2) transpositions. 
    So the sign of the permutation which moves a1 to the front of the list 
    yet keeping the order all other elements unchanged is negative iff pos is odd.
    Then, as the sign of a composition of two permutations pi1 o pi2 
    is the product of of the signs of each permutation the result is the 
    sign of the transposition of a1 to the first place multiplied 
    with the sign of the permutation of the remaining elements fixing a1. 
    But this is the same as the sign of the permutation of the rests 
    of the lists. 
    But transposing an element towards the first place and then taking 
    the rest is the same as simply removing the element.
    Hence sign(l1 l2) for non-empty l1 and l2 is the same as 
    (-1)^(position (first l1) l2) * (sign (rest l1) (remove (first l1) l2)).
    This is done in the do-list, where the sign of the result is toggled 
    iff the position is odd.
|#


;;; Simplicial boundaries:
;;; An n-simplex is a list of (distinct) vertices. A vertex may be any object.
;;; A chain is the formal sum ca * A + cb * B + ... of elements A, B, ... with 
;;; (integer) coefficients ca, cb, ... . These are expressed as an a-list
;;; ((A . ca) (B . cb) ... ).

(defun simplicial-boundary (simplex)
  "Returns the simplicial boundary of the specified simplex as 
a chain of simplexes."
  (let ((sign +1)
	(result ()))
    (dotimes (i (length simplex) (nreverse result))
      (push `(,(remove-nth i simplex) . ,sign)
	    result)
      (setf sign (- sign)))))


(defun simplicial-boundary-row (simplex)
  "Returns the simplicial boundary of the specified simplex as 
a matrix row with the specified simplex as row-index."
  (let ((sign +1)
	(result ()))
    (dotimes (i (length simplex) (nreverse result))
      (push `((,simplex ,(remove-nth i simplex)) . ,sign)
	    result)
      (setf sign (- sign)))))


(defun simplicial-chain-boundary (simplexes)
  "Returns the simplicial boundary of a chain of simplexes as a chain."
  (let ((result ()))
    (dolist (simplex simplexes result)
      (setf result 
	    (chain+ (chain* (cdr simplex)
			    (simplicial-boundary (car simplex)))
		    result
		    :test #'set-equals)))))



;;; obsolete "old-style" closure and boundary functios.
(obsolete get-co-closure get-closure*)
(obsolete get-closure    get-closure*)
(obsolete get-co-star    get-star*   )
(obsolete get-star       get-star*   )

(defun get-co-closure (cells co-boundaries &key (test #'eql))
  
  "Returns the topological boundary of a set of n-cells as an reversed n-1-complex.  

The resulting boundary operators are returned as primary value and the cells 
found thereby as secondary value."

  (let ((padded-co-cellss 
	 (cons (list cells)
	       (mapcar (constantly ())
		       co-boundaries))))
    (multiple-value-bind (bd clss)
	(get-closure* (nreverse padded-co-cellss)
		      (reverse co-boundaries)
		      :test test)
      (values (nreverse bd) (nreverse clss)))))


(defun get-closure (cells boundaries &key (test #'eql))
  "Returns the topological closure of a set of n-cells in an n-complex which has 
the specified boundary.

The resulting boundary operators are returned as primary value and the cells 
found thereby as secondary value."

  (let ((padded-cellss 
	 (nreverse (cons cells
			 (mapcar (constantly ()) boundaries)))))
	(get-closure* padded-cellss boundaries :test test)))


(defun get-co-star (co-cells co-boundaries &key (test #'eql))
  "Returns the topological star neighbourhood of a set of vertices in an n-complex which has 
the specified co-boundary.

The resulting co-boundary operators are returned as primary value and the co-cells 
found thereby as secondary value."

  (let ((padded-co-cellss 
	 (cons co-cells
	       (mapcar (constantly ()) co-boundaries))))
    (multiple-value-bind (bd clss)
	(get-star* (nreverse padded-co-cellss)
		   (reverse co-boundaries)
		   :test test)
      (values (nreverse bd) (nreverse clss)))))


(defun get-star (cells boundaries &key (test #'eql))
  "Returns the topological star eighbourhood of a set of vertices as a complex.  

The resulting boundary operators are returned as primary value and the cells 
found thereby as secondary value."

  (get-star* (list cells) boundaries :test test))


;; The new star and boundary functions:

(defun get-star* (cellss boundaries &key (test #'eql))
  "Returns the topological star eighbourhood of a sequence of sets of cells 
\(hence cellss - a plural of a plural) as a complex.  

The resulting boundary operators are returned as primary value and the cells 
found thereby as secondary value.

The top level cons chains of the resulting lists are newly allocated.
Each entry in such list may share structure with the input data."

  (do* ((subtract-op       (rcurry #'set-difference    :test test))
	(add-op            (rcurry #'union             :test test))
	(matrix-test       #'(lambda (a b)
			       (or (eq a b)
				   (and (funcall test 
						 (matrix-entry-row a)
						 (matrix-entry-row b))
					(funcall test 
						 (matrix-entry-col a)
						 (matrix-entry-col b))))))
	(add-bd-op         (rcurry #'union  :test matrix-test))
	(to-set-op         (rcurry #'remove-duplicates :test test))

	;; collectors for cells and boundaries: one per dimension.
	(result-cellss     (cons () (mapcar (constantly ()) boundaries)))
	(result-boundaries (mapcar (constantly ()) boundaries))

	;; those of the next cells which aren't yet in the result
	;; (initially all)
	(new-cellss        (map-into (copy-list result-cellss) ; pad cells
				     #'identity
				     cellss))
	(new-boundaries    (mapcar #'(lambda (boundary cells)
				       (mapcan #'(lambda (cell)
						   (matrix-get-col 
						    boundary cell 
						    :test test))
					       cells))
				   boundaries
				   new-cellss))
	;; helper: next star cells found during iteration
	(next-cellss       ()))

    ((every #'null new-cellss)
     (values result-boundaries
	     result-cellss))

    (setf result-cellss     (map-into result-cellss add-op        ; collect
				      result-cellss
				      new-cellss)
	  result-boundaries (map-into result-boundaries add-bd-op ; collect
				      result-boundaries
				      new-boundaries)

	  ;; do the next step with boundaries:
	  new-boundaries    (map-into new-boundaries 
				      #'(lambda (boundary cells)
					  (mapcan #'(lambda (cell)
						      (matrix-get-col boundary cell 
								      :test test))
						  cells))
				      boundaries
				      new-cellss)
	  ;; the next boundary cells...
	  next-cellss       (cons () ; do one dimension step
				  (mapcar 
				   to-set-op ; convert to sets
				   (mapcar #'(lambda (boundary)
					       (mapcar #'matrix-entry-row 
						       boundary))
					   new-boundaries)))
	  ;; ... if not yet collected.
	  new-cellss        	  (mapcar subtract-op
					  next-cellss
					  result-cellss))))


(defun get-closure* (cellss boundaries &key (test #'eql))
  "Returns the topological closure of a sequence of sets of cells 
\(hence cellss - a plural of a plural) as a complex.  

The resulting boundary operators are returned as primary value and the cells 
found thereby as secondary value.

The top level cons chains of the resulting lists are newly allocated.
Each entry in such list may share structure with the input data.
"

  (do* ((subtract-op       (rcurry #'set-difference    :test test))
	(add-op            (rcurry #'union             :test test))
	(matrix-test       #'(lambda (a b)
			       (or (eq a b)
				   (and (funcall test 
						 (matrix-entry-row a)
						 (matrix-entry-row b))
					(funcall test 
						 (matrix-entry-col a)
						 (matrix-entry-col b))))))
	(add-bd-op         (rcurry #'union  :test matrix-test))
	(to-set-op         (rcurry #'remove-duplicates :test test))

	;; collectors for cells and boundaries: one per dimension.
	(result-cellss     (cons () (mapcar (constantly ()) boundaries)))
	(result-boundaries (mapcar (constantly ()) boundaries))

	;; those of the next cells which aren't yet in the result
	;; (initially all cellss)
	(new-cellss        (map-into (copy-list result-cellss) ; pad cells.
				     #'identity
				     cellss))
	(new-boundaries    (mapcar #'(lambda (boundary cells)
				       (mapcan #'(lambda (cell)
						   (matrix-get-row 
						    boundary cell 
						    :test test))
					       cells))
				   boundaries
				   (cdr new-cellss)))
	;; helper: next star cells found during iteration
	(next-cellss       ()))

    ((every #'null new-cellss)
     (values result-boundaries
	     result-cellss))

    (setf result-cellss     (map-into result-cellss add-op        ; collect
				      result-cellss
				      new-cellss)
	  result-boundaries (map-into result-boundaries add-bd-op ; collect
				      result-boundaries
				      new-boundaries)

	  ;; do the next step with boundaries:
	  new-boundaries    (map-into new-boundaries 
				      #'(lambda (boundary cells)
					  (mapcan #'(lambda (cell)
						      (matrix-get-row 
						       boundary 
						       cell
						       :test test))
						  cells))
				      boundaries
				      (cdr new-cellss))
	  ;; the next boundary cells...
	  next-cellss       (mapcar to-set-op ; convert to sets
				    (mapcar #'(lambda (boundary)
						(mapcar #'matrix-entry-col 
							boundary))
					    new-boundaries))
	  ;; ... if not yet collected.
	  new-cellss        	  (mapcar subtract-op
					  next-cellss
					  result-cellss))))



(defun complex-co-delete-cell (dropee co-boundary co-cells &key (test #'eql))
  (let* ((dropee-pos (position dropee (car co-cells) :test test))
	 (co-bd      co-boundary)
	 (co-ce      co-cells))
    (when dropee-pos
	(setf (matrix-get-row (car co-bd) dropee :test test) #| to |#    ()
	      (car co-ce)     #| to |#  (delete-nth dropee-pos (car co-ce))))
    (values co-bd co-ce)))



(defun complex-delete-cell (dropee boundary cells &key (test #'eql))
  (multiple-value-bind (co-bd co-ce)
      (complex-co-delete-cell dropee (nreverse boundary) (nreverse cells) :test test)
    (values (nreverse co-bd) 
	    (nreverse co-ce))))


(defun complex-co-remove-cell (dropee co-boundary co-cells &key (test #'eql))
  (let* ((dropee-pos (position dropee (car  co-cells) :test test)))
    (if dropee-pos
	;; changed:
	(values 
	 (cons (matrix-set-row (car co-boundary) dropee
			       ()
			       :test test)
	       (cdr co-boundary))
	 (cons (remove-nth dropee-pos (car co-cells))
	       (cdr co-cells)))
	;; unchanged:
	(values co-boundary 
		co-cells))))



(defun complex-remove-cell (dropee cells boundary &key (test #'eql))
  (multiple-value-bind (co-boundary co-cells)
      (complex-co-remove-cell dropee (reverse cells) (reverse boundary) :test test)
    (values (nreverse co-boundary) 
	    (nreverse co-cells))))



(defun-with-default-output-keys write-boundary 
    (matrix &key (test #'eql) (symbol "d") (dimension nil))

  "A pretty printer for boundary matrices."

  (let* ((cells ()))
    (dolist (entry matrix)
      (pushnew (matrix-entry-row entry) cells :test test))
    (dolist (row-idx (nreverse cells))
      (if dimension
	  (format T "~%~A[~A](~W) = " symbol dimension row-idx)
	  (format T "~%~A(~W) = " symbol row-idx))
      (let ((row (matrix-get-row matrix row-idx :test test))
	    (plus ""))
	(dolist (entry row)
	  (let ((value (matrix-entry-value entry)))
	    (if (minusp value)
		(format T " - ")
		(format T plus))
	    (setf plus " + ")
	    (if (= 1 (abs value))
		(format T "~W"
			(matrix-entry-col entry))
		(format T "~W*~W"
			(abs (matrix-entry-value entry))
			(matrix-entry-col entry))))))))
  matrix )

(defun write-boundaries  (matrices &rest keys)

  (let ((dim  (getf keys :dimension))
	(keys (copy-list keys)))
    (remf keys :dimenion)

    (if dim
	(dolist (matrix matrices)
	  (apply #'write-boundary matrix :dimension dim keys)
	  (incf dim))

	(dolist (matrix matrices)
	  (apply #'write-boundary matrix keys)))

    matrices))


(defun triangulate-cell (apex cell boundary cells &key (test #'eql)
			 (tensor-marker '#:\(x\))
			 (make-tensor #'(lambda (a b) (list* tensor-marker a b)))
			 ;;(tensor-pi-1 #'(lambda (axb) (cadr axb)))
			 (tensor-pi-2 #'(lambda (axb) (cddr axb))))

  (multiple-value-bind (co-boundary co-cells)
      (co-triangulate-cell apex cell (reverse boundary) (reverse cells) 
			   :test test
			    :tensor-marker tensor-marker
			    :make-tensor make-tensor
			    ;(tensor-pi-1 #'(lambda (axb) (cadr axb)))
			    :tensor-pi-2 tensor-pi-2)
    (values (nreverse co-boundary) 
	    (nreverse co-cells))))

(defun-with-default-output-keys write-set (seq &key (stream *standard-output*))
  (let ((format-stream (if (null stream) t stream)))
    (format format-stream "{")
  (when seq
    (write (car seq) :stream stream))
  (when (cdr seq)
    (dolist (item (cdr seq))
      (format format-stream ", ~W" item)))
    (format format-stream "}"))
  seq )


(defun pack-boundary (boundaries &key (test #'eql) (row-test test) (col-test test))

  "Adds multiple occurrences of all boundary matrices into one occurrence."

  (let ((entry-test (matrix-key-test row-test col-test)))
    (mapcar #'(lambda (boundary)
		(reduce (rcurry #'chain+ :test entry-test)
			(mapcar #'list boundary)
			:initial-value ()))
	    boundaries)))


(defun extend-complex (boundaries cellss &optional (negative-dim-cell ()))

  "Extends the complex specified by boundaries and cellss by a negative-dimensional 
singleton set (<>) of exactly one cell <>, the (-1)-cell. Each null-cell x has then 
the extended boundary d(x) = <>. 

It is not guaranteed that the resultinmg boundary satisfies boundaryp, as the 
cell <> is considered 0, in theory, but can occur with coefficent different 
to 0. 
Hence 
  d(d(x)) = some-coefficient * <> = some-coefficient * 0 = 0 
in theory, but boundaryp only checks the left-hand side: 
\(zerop some-coefficient), which may not be true."

  (values `(,(mapcar #'(lambda (cell) 
			 (make-matrix-entry cell negative-dim-cell 1))
		     (car cellss))
	     . ,boundaries)
	  `((,negative-dim-cell)  . ,cellss)))


(defmacro comment (&rest body) 
  "Simply removes the body and replaces it by nil."
  (declare (ignore body))
  ())


(defun verbose-co-triangulate-cell (apex cell co-boundary co-cells &key (test #'eql)
			    (verbose T) ; print debug information.
			    (tensor-marker '#:\(x\))
			    (make-tensor #'(lambda (a b) (list* tensor-marker a b)))
			    ;(tensor-pi-1 #'(lambda (axb) (cadr axb)))
			    (tensor-pi-2 #'(lambda (axb) (cddr axb)))
			    &aux #|(sim-test (curry #'simplex-test test))|#
			    (tree-test (curry #'tree-test test)))

  "Combinatorically triangulates the closure of the specified n-cell in the complex defined 
by the reversed sequence of boundaries co-boundary and the reversed sequence of cell sets 
co-cells. 

The triangulation is done by consing the specified apex p to each boundary cell c of the 
specified cell which gives a new cell (p . c). These cells are added to the 
boundary complex of cell as i+1-cell if c is an i-cell. 
Additionally (p) is added as  avertex to the new complex.

The boundary d(p . c) of each new cell (p . c) is then c - p x d(c). 
p x d(c) is consing p to each cell z in the linear combination of d(c):
  If d(c) = a * A + b * B + ...
then 
  p x d(c) = a * (p . A) + b * (p . B) + ...
Thereby the specified cell is replaced by a linear combination 
  (p . A) + (p . B) + (p . C) + ...
where A, B, C ... are the boundary cells of C. Note that this (still) only works when 
p is not a vertex of the specified complex."
  ;; Remember: co-<foo> are <foo> in reversed order:
  ;; co-cells are (C_n C_n-1 ... C_0)
  ;; co-boundary is (d_n d_n-1 ... d_1)
  ;; for a complex :
  ;;           d_n             d_n-1                     d_1
  ;;   C_n  --------> C_n-1 ----------> C_n-2 --> ... --------> C_0
  ;;
  ;; what is to construct is:
  ;;
  ;;                              d_n-1            d_n-2          d_1
  ;;                     C_n-1 ----------> C_n-2 ------> ... --------> C_0
  ;;                   ,;                 ,;                           ,;
  ;;                  ´/'                ´/'                          ´/'
  ;;                  /                  /                            /
  ;;                 /                  /                            /
  ;;                /                  /                            /
  ;;          (+/-pi_n-1)        (+pi_n-2)         ...          (+pi_0)
  ;;             /                   /                            /        
  ;;            /                   /                            /        
  ;;           /                   /                            /        
  ;;          /                   /                            /        
  ;;         /                   /                            /        
  ;;  pxC_n-1 ----------> pxC_n-2 -------->...--------> pxC_0  -------> (p)
  ;;        (-/+ pxd_n-1)         -pxd_n-2     -pxd_1              -p
  ;; 
  ;; Where the signs +/- for pi_n-1 and -/+ for pxd_n-1 are taken from d_n.


  (let (closure-boundaries           co-closure-boundaries 
	closure-cells 	             co-closure-cells 
	#|first-boundary|#           first-co-boundary 
	#|rest-boundaries|#          rest-co-boundaries 
	#|tensor-product-cells|#     co-tensor-projection 
	#|tensor-product-boundary|#  co-tensor-product-boundaries 
				     initial-co-boundary-p
				     cellss 
				     boundary)
    (unless (setf initial-co-boundary-p (co-boundaryp co-boundary :test test))
      (warn "initial \"co-boundary\" is not a boundary operator"))

    (when verbose
      (let ((dim (1- (length co-cells))))
	(format T "~&The complex is")
	(when co-cells 
	  (format T "~&  C[~S]" dim))
	(dotimes (n-i dim)
	  (let ((i (- dim n-i 1)))
	    (format T " ---d[~S]--> C[~S]" i i)))
	(format T "~&with cells:")
	(let ((dim dim))
	(dolist (c-i co-cells)
	  (format T "~&  C[~S] = Z" dim) 
	  (write-set c-i :escape nil :readably nil)
	  (decf dim)))
	(let ((dim dim))
	  (format T "~&and with boundary:")
	  (dolist (bd co-boundary)
	    (write-boundary bd :symbol "  d" :dimension (decf dim) :escape nil :readably nil)))
	(if initial-co-boundary-p
	    (format T "~&a boundary, indeed.")
	    (format T "~&but it is NOT a boundary!"))))

    (setf ;; padding cells into a sequence of cell-sets :
          cellss                   `(,@(mapcar (constantly ()) co-boundary)  (,cell)) ; 0 ... n
	  ;; reversing co-boundary: 1->0 ... n-1->n
	  boundary                  (reverse co-boundary)
	  ;; extended complex of the cell boundary: -1 ... n and 
	  (values closure-boundaries ; -1 ... n
		  closure-cells)     ; 0->(-1) ... n-1->n
	                            (multiple-value-call #'extend-complex
				      (get-closure* cellss  boundary :test test))
	  co-closure-boundaries     (reverse closure-boundaries) 
	  co-closure-cells          (reverse closure-cells)      
          first-co-boundary         (car co-closure-boundaries)  ; n --> n-1  gets "merged"
          rest-co-boundaries        (cdr co-closure-boundaries)) ; rest  n-1 --> ... --> -1 will be "returned"
    
    (when verbose
      (let ((*print-escape* nil) (*print-readably* nil)
	    (dimension (length co-boundary)))
	(format T "~&The extended cell closure of ~W is:" cell)
	(dolist (bd co-closure-boundaries)
	  (write-boundary bd :symbol "  d" :dimension (decf dimension)))))

    ;; establish a mapping from a tensor-product cell to the original cell. 
    ;;   This is the "shift-matrix" pi: d(p x C) = C. (simply a projection)
    (setf co-tensor-projection   ; n ... 0
	  (mapcar #'(lambda (cells)
		      (mapcar #'(lambda (cell)
				  (let ((apex-x-cell (funcall make-tensor apex cell)))
				    (make-matrix-entry  apex-x-cell  cell 1)))
			      cells))
		  (cdr co-closure-cells)))       ; start with n-1: (n-1  n-2  ... 2  1   0  -1)
    ;; now we have the the tensor product cells shifted from (n-1 ... -1) to (n ... 0)
    ;; by (n  n-1 ... 1 0) is apex x (n-1 n-2 ... 0 -1).

    (when verbose
      (format t "~&The projection (pasting) boundary part:")
      (let ((dim (length co-cells)))
	(mapc #'(lambda (bd-i) (write-boundary bd-i :symbol "  pi" :dimension (decf dim) :escape nil :readably nil))
	      co-tensor-projection))
      (format t "~&Computing the new triangle boundary part (apex x d):"))

    ;; The (negative) copy of the complex boundary tagged by apex:
    (setf co-tensor-product-boundaries
	  (mapcar #'(lambda (boundary row-cells col-cells)
		      (mapcar #'(lambda (entry)
				  (let ((row-idx  (matrix-entry-row entry))
					(col-idx  (matrix-entry-col entry))
					(-value   (- (matrix-entry-value entry)))
					new-row new-col result)
				    ;; make the new matrix-entry
				    (setf new-row (matrix-entry-row (find  row-idx row-cells 
								    :test test
								    :key  #'matrix-entry-col))
					  new-col (matrix-entry-row (find col-idx col-cells 
								   :test test
								   :key #'matrix-entry-col))
					  result  (make-matrix-entry new-row new-col -value))
				  (when verbose 
				    (format t "~%shifting ~A to ~A " entry result))
				  (when (member new-col row-cells 
						:test tree-test
						:key  #'matrix-entry-col)
				    (warn "Cell duplication: ~A!" new-col))
				  result))
			      boundary ))
		  rest-co-boundaries                 ; n-1 -> n-2 ... 
		  co-tensor-projection            ; n   -->  n-1  n-1 --> n-2 ...   0
		  (cdr co-tensor-projection)))    ; n-1 -->  n-2  n-2 --> n-3 ... (-1)

    (when verbose
      (format t "~%The shifted boundary:")
      (let ((dim (length co-cells)))
	(mapc #'(lambda (bd-i)
		  (write-boundary bd-i :symbol "  d*" :dimension (decf dim) :escape nil :readably nil))
	      co-tensor-product-boundaries)))

    ;; Now we must get rid of degenerate cells by adding them to their "buddies":
    (setf co-tensor-product-boundaries 
	  ;; this is dummy-code yet:
	  (map-into co-tensor-product-boundaries 
		    #'(lambda(boundary n-cells n-1-cells) boundary)
		    co-tensor-product-boundaries
		    co-tensor-projection
		    (cdr co-tensor-projection)))


    ;; Now we have the three ingredients and could add them already. 
    ;; However, we want to correct the signs of the first boundaries: 
  
    ;; Correct the first of co-tensor-product cells ...
    (when verbose
      (format T "~&correcting shifted boundary:"))

    (dolist (entry (car co-tensor-projection))
      (when verbose (format T "~&  ~A " entry))
      (let ((bd-cell (matrix-entry-col entry)))
	(setf (cdr entry) 
	      (* (cdr entry) 
		 (cdr (find bd-cell first-co-boundary
			    :test test
			    :key  #'matrix-entry-col)))))
      (when verbose (format T " becomes ~A." entry)))

    ;; and of co-tensor-product-boundaries:
    (when verbose
      (format T "~&correcting pasting boundary:"))
    (dolist (entry (car co-tensor-product-boundaries))
      ;; entry is ((pxc_n-1 pxc_n-2) . D(c_n-1 c_n-2))
      (when verbose
	(format T "~&  ~A " entry))
      (let ((bd-cell (funcall tensor-pi-2 (matrix-entry-row entry)))) ; the projection pi(p x C) +-> C
	(setf (cdr entry) 
	      (* (cdr entry) 
		 (cdr (find  bd-cell first-co-boundary
			     :test test
			     :key  #'matrix-entry-col)))))
      (when verbose (format T " becomes ~A." entry)))

    (values 
     ;; the primary value is the new boundary operator:
      
     (let* ((dim (length co-boundary))
	    (result-co-boundary 
	     (mapcar #'(lambda(top  diag  bottom &aux sum)
			 (when verbose 
			   (let ((*print-escape*  nil) (*print-readably* nil))
			     (format T "~&adding")
			     (write-boundary top :symbol    "      d" :dimension dim)
			     (write-boundary diag :symbol   "+   -d*" :dimension dim)
			     (write-boundary bottom :symbol (format nil "+   ~A.d" apex)
					     :dimension dim)
			     (format T "~& = ")))
			 (setf sum (chain+ (chain+ top diag :test tree-test) bottom :test tree-test))
			 (when verbose 
			   (write-boundary sum :symbol "  t" :dimension dim :escape nil :readably nil))
			 (decf dim)
			 sum)
		    (cons () rest-co-boundaries)    ; the top row 
		    co-tensor-projection            ; the diagonal "backmaps"
		    co-tensor-product-boundaries))) ; the bottom-row 
       (when (and initial-co-boundary-p
		  (not (co-boundaryp result-co-boundary :test tree-test)))
	 (warn "OOPS!! Now I have screwed up a co-boundary to:~A~&(sorry for that! *hm* *hm* *he* *he*!)" 
	       ;; This can happen when a pair of cells
	       ;; (make-tensor apex k) and k exist in the input complex.
	       (let ((*print-escape* nil) 
		     (*print-readably* nil) 
		     (*standard-output* (make-string-output-stream))
		     (dim (length co-cells)))
		 (mapc #'(lambda (bd-i) (write-boundary bd-i :symbol "  d*" :dimension (decf dim)))
		       result-co-boundary)
		 (get-output-stream-string *standard-output*))))

       result-co-boundary) ; the actual value
		   

     ;; the secondary value is the sequence of the new tensor-product-cells:
     (mapcar #'(lambda (diag)
		 (delete-duplicates (mapcar #'matrix-entry-row diag)
				    :test tree-test))
	     co-tensor-projection))))


(defun co-triangulate-cell (apex cell co-boundary co-cells &key (test #'eql)
			    (tensor-marker '#:\(x\))
			    (make-tensor #'(lambda (a b) (list* tensor-marker a b)))
			    ;(tensor-pi-1 #'(lambda (axb) (cadr axb)))
			    (tensor-pi-2 #'(lambda (axb) (cddr axb)))
			    &aux #|(sim-test (curry #'simplex-test test))|#
			    (tree-test (curry #'tree-test test)))

  "Combinatorically triangulates the closure of the specified n-cell in the complex defined 
by the reversed sequence of boundaries co-boundary and the reversed sequence of cell sets 
co-cells. 

The triangulation is done by consing the specified apex p to each boundary cell c of the 
specified cell which gives a new cell (p . c). These cells are added to the 
boundary complex of cell as i+1-cell if c is an i-cell. 
Additionally (p) is added as  avertex to the new complex.

The boundary d(p . c) of each new cell (p . c) is then c - p x d(c). 
p x d(c) is consing p to each cell z in the linear combination of d(c):
  If d(c) = a * A + b * B + ...
then 
  p x d(c) = a * (p . A) + b * (p . B) + ...
Thereby the specified cell is replaced by a linear combination 
  (p . A) + (p . B) + (p . C) + ...
where A, B, C ... are the boundary cells of C. Note that this (still) only works when 
p is not a vertex of the specified complex."
  ;; Remember: co-<foo> are <foo> in reversed order:
  ;; co-cells are (C_n C_n-1 ... C_0)
  ;; co-boundary is (d_n d_n-1 ... d_1)
  ;; for a complex :
  ;;           d_n             d_n-1                     d_1
  ;;   C_n  --------> C_n-1 ----------> C_n-2 --> ... --------> C_0
  ;;
  ;; what is to construct is:
  ;;
  ;;                              d_n-1            d_n-2          d_1
  ;;                     C_n-1 ----------> C_n-2 ------> ... --------> C_0
  ;;                   ,;                 ,;                           ,;
  ;;                  ´/'                ´/'                          ´/'
  ;;                  /                  /                            /
  ;;                 /                  /                            /
  ;;                /                  /                            /
  ;;          (+/-pi_n-1)        (+pi_n-2)         ...          (+pi_0)
  ;;             /                   /                            /        
  ;;            /                   /                            /        
  ;;           /                   /                            /        
  ;;          /                   /                            /        
  ;;         /                   /                            /        
  ;;  pxC_n-1 ----------> pxC_n-2 -------->...--------> pxC_0  -------> (p)
  ;;        (-/+ pxd_n-1)         -pxd_n-2     -pxd_1              -p
  ;; 
  ;; Where the signs +/- for pi_n-1 and -/+ for pxd_n-1 are taken from d_n.


  (let (closure-boundaries           co-closure-boundaries 
	closure-cells 	             co-closure-cells 
	#|first-boundary|#           first-co-boundary 
	#|rest-boundaries|#          rest-co-boundaries 
	#|tensor-product-cells|#     co-tensor-projection 
	#|tensor-product-boundary|#  co-tensor-product-boundaries 
				     initial-co-boundary-p
				     cellss 
				     boundary)
    (unless (setf initial-co-boundary-p (co-boundaryp co-boundary :test test))
      (warn "initial \"co-boundary\" is not a boundary operator"))

    (setf ;; padding cells into a sequence of cell-sets :
          cellss                   `(,@(mapcar (constantly ()) co-boundary)  (,cell)) ; 0 ... n
	  ;; reversing co-boundary: 1->0 ... n-1->n
	  boundary                  (reverse co-boundary)
	  ;; extended complex of the cell boundary: -1 ... n and 
	  (values closure-boundaries ; -1 ... n
		  closure-cells)     ; 0->(-1) ... n-1->n
	                            (multiple-value-call #'extend-complex
				      (get-closure* cellss  boundary :test test))
	  co-closure-boundaries     (reverse closure-boundaries) 
	  co-closure-cells          (reverse closure-cells)      
          first-co-boundary         (car co-closure-boundaries)  ; n --> n-1  gets "merged"
          rest-co-boundaries        (cdr co-closure-boundaries)) ; rest  n-1 --> ... --> -1 will be "returned"
    
    ;; establish a mapping from a tensor-product cell to the original cell. 
    ;;   This is the "shift-matrix" pi: d(p x C) = C. (simply a projection)
    (setf co-tensor-projection   ; n ... 0
	  (mapcar #'(lambda (cells)
		      (mapcar #'(lambda (cell)
				  (let ((apex-x-cell (funcall make-tensor apex cell)))
				    (make-matrix-entry  apex-x-cell  cell 1)))
			      cells))
		  (cdr co-closure-cells)))       ; start with n-1: (n-1  n-2  ... 2  1   0  -1)
    ;; now we have the the tensor product cells shifted from (n-1 ... -1) to (n ... 0)
    ;; by (n  n-1 ... 1 0) is apex x (n-1 n-2 ... 0 -1).

    ;; The (negative) copy of the complex boundary tagged by apex:
    (setf co-tensor-product-boundaries
	  (mapcar #'(lambda (boundary row-cells col-cells)
		      (mapcar #'(lambda (entry)
				  (let ((row-idx  (matrix-entry-row entry))
					(col-idx  (matrix-entry-col entry))
					(-value   (- (matrix-entry-value entry)))
					new-row new-col result)
				    ;; make the new matrix-entry
				    (setf new-row (matrix-entry-row (find  row-idx row-cells 
								    :test test
								    :key  #'matrix-entry-col))
					  new-col (matrix-entry-row (find col-idx col-cells 
								   :test test
								   :key #'matrix-entry-col))
					  result  (make-matrix-entry new-row new-col -value))
				  (when (member new-col row-cells 
						:test tree-test
						:key  #'matrix-entry-col)
				    (warn "Cell duplication: ~A!" new-col))
				  result))
			      boundary ))
		  rest-co-boundaries                 ; n-1 -> n-2 ... 
		  co-tensor-projection            ; n   -->  n-1  n-1 --> n-2 ...   0
		  (cdr co-tensor-projection)))    ; n-1 -->  n-2  n-2 --> n-3 ... (-1)

    ;; Now we must get rid of degenerate cells by adding them to their "buddies":
    (setf co-tensor-product-boundaries 
	  ;; this is dummy-code yet:
	  (map-into co-tensor-product-boundaries 
		    #'(lambda(boundary n-cells n-1-cells) boundary)
		    co-tensor-product-boundaries
		    co-tensor-projection
		    (cdr co-tensor-projection)))


    ;; Now we have the three ingredients and could add them already. 
    ;; However, we want to correct the signs of the first boundaries: 
  
    ;; Correct the first of co-tensor-product cells ...
    (dolist (entry (car co-tensor-projection))
      (let ((bd-cell (matrix-entry-col entry)))
	(setf (cdr entry) 
	      (* (cdr entry) 
		 (cdr (find bd-cell first-co-boundary
			    :test test
			    :key  #'matrix-entry-col))))))

    ;; and of co-tensor-product-boundaries:
    (dolist (entry (car co-tensor-product-boundaries))
      ;; entry is ((pxc_n-1 pxc_n-2) . D(c_n-1 c_n-2))
      (let ((bd-cell (funcall tensor-pi-2 (matrix-entry-row entry)))) ; the projection pi(p x C) +-> C
	(setf (cdr entry) 
	      (* (cdr entry) 
		 (cdr (find  bd-cell first-co-boundary
			     :test test
			     :key  #'matrix-entry-col))))))

    (values 
     ;; the primary value is the new boundary operator:
      
     (let* ((result-co-boundary 
	     (mapcar #'(lambda(top  diag  bottom &aux sum)
			 (setf sum (chain+ (chain+ top diag :test tree-test) bottom :test tree-test))
			 sum)
		    (cons () rest-co-boundaries)    ; the top row 
		    co-tensor-projection            ; the diagonal "backmaps"
		    co-tensor-product-boundaries))) ; the bottom-row 
       (when (and initial-co-boundary-p
		  (not (co-boundaryp result-co-boundary :test tree-test)))
	 (warn "OOPS!! Now I have screwed up a co-boundary to:~A~&(sorry for that! *hm* *hm* *he* *he*!)" 
	       ;; This can happen when a pair of cells
	       ;; (make-tensor apex k) and k exist in the input complex.
	       (let ((*print-escape* nil) 
		     (*print-readably* nil) 
		     (*standard-output* (make-string-output-stream))
		     (dim (length co-cells)))
		 (mapc #'(lambda (bd-i) (write-boundary bd-i :symbol "  d*" :dimension (decf dim)))
		       result-co-boundary)
		 (get-output-stream-string *standard-output*))))

       result-co-boundary) ; the actual value
		   

     ;; the secondary value is the sequence of the new tensor-product-cells:
     (mapcar #'(lambda (diag)
		 (delete-duplicates (mapcar #'matrix-entry-row diag)
				    :test tree-test))
	     co-tensor-projection))))



(defun map-complex (func boundaries cellss &key (test #'eql) (keep-mapping nil))

  "Applies the function func to each cell of the specified complex, adjusts 
the boundaries accordingly and returns the new complex. The returnd 
boundaries and cell sets have the same cons-tree-structure as the original
complex.

If the keep-mapping keyword argument is true, the cells will be returned as 
pairs (old-cell . (func old-cell)) where new-cell = (funcall func old-cell)).
"

  (let* ((tagged-cellss 
	  (mapcar #'(lambda (cells)
		      (mapcar #'(lambda (cell)
				  (cons cell (funcall func cell)))
			      cells))
		  cellss))
	(tagged-boundaries
	 (mapcar #'(lambda (boundary row-cells col-cells)
		     (mapcar #'(lambda (matrix-entry)
				 (let ((old-row (matrix-entry-row matrix-entry))
				       (old-col (matrix-entry-col matrix-entry))
				       (value   (matrix-entry-value matrix-entry)))
				   (let ((new-row (assoc old-row row-cells :test test))
					 (new-col (assoc old-col col-cells :test test)))
				     (assert new-row)
				     (assert new-col)
				     (make-matrix-entry (cdr new-row) (cdr new-col) value))))
			     boundary))
		 boundaries
		 (cdr tagged-cellss)
		 tagged-cellss)))
		
    (values tagged-boundaries

	    (if keep-mapping
		tagged-cellss
		(map-into tagged-cellss
			  #'(lambda (tagged-cells)
			      (map-into tagged-cells 
					#'cdr tagged-cells))
			  tagged-cellss)))))




(defun tag-complex (boundaries cellss apex &key (tensor #'cons) (test #'eql) (keep-mapping nil))

  "Applies (funcall tensor apex cell) to each cell of the specified complex, adjusts 
the boundaries accordingly and returns the new complex."

  (map-complex (curry tensor apex) boundaries cellss :test test :keep-mapping keep-mapping))



(defun triangulate-cell-2 (cell boundaries cellss &key (apex (gensym)) (test #'eql))

  "Triangulates the n-cell in the specified n-dimensional complex by removing that 
cell and topologically joining the apex as a new vertex with that cell's boundary. 

The result will not be a complex if apex is a vertex of the cell's boundary."

  (declare (ignorable cellss))
  
  (let (closure-cellss
	closure-boundaries
	closure-n-1-boundary
	abs-closure-n-2-boundary

	cone-cellss 
	cone-boundaries
	cone-n-cells

	paste-boundaries
	paste-n-1-boundary)

    ;; compute the cell closure complex:
    (setf (values closure-boundaries  ; 0 <- 1 ... n-1<-n
		  closure-cellss)     ; 0 ... n
	  (get-closure* (nconc (mapcar (constantly ()) boundaries)  ; pad cell
			       `((,cell)))
			boundaries
			:test test))

    ;; The last closure cell set is not interesting.
    ;; The coefficients of that cells boundary, however, are used.
    (setf closure-n-1-boundary  (first (last closure-boundaries))) ; n-1 <- n
    (setf (values closure-boundaries                        ; -1 <- 0 <- 1 ... n-2 <- n-1 
		  closure-cellss)                           ; -1 0 ... n-1
	  (extend-complex (butlast closure-boundaries)      ; 0 <- 1 ... n-2 <- n-1
			  (butlast closure-cellss)))        ; 0 ... n-1
    ;; now the closure is extended

    ;; pre-correct the last closure boundary (before shifting):
    (setf abs-closure-n-2-boundary
	  (mapcar #'(lambda (entry)
		      (let* ((cell-bd (matrix-entry-row entry))
			     (factor  (matrix-get closure-n-1-boundary 
						  cell cell-bd 
						  :test test)))
			(format t "~&  factor: ~A" factor)
			(if (= factor 1)
			    entry
			    (make-matrix-entry cell-bd
					       (matrix-entry-col entry)
					       (* factor 
						  (matrix-entry-value entry))))))
		  (first (last closure-boundaries))))

    ;; compute the cone-cell - base-cell pairs and 
    ;;   the boundary that pastes cone-cells to cone-cells.
    ;;   This generates the new cells which are to replace the specified cell:
    (setf (values cone-boundaries    ; 0 <- 1 ... n-1 <- n
		  cone-cellss)       ; (-1 0) ... (n-2 n-1) (n-1 n)    ; the cells and kept mappings
	  (tag-complex (nconc (butlast closure-boundaries)
			      (list abs-closure-n-2-boundary))
		       closure-cellss
		       apex 
		       :tensor #'cons 
		       :test   test 
		       :keep-mapping t)
	  cone-n-cells (first (last cone-cellss)))

    ;; compute the boundary that pastes cone-cells to base-cells:
    (setf paste-n-1-boundary ; treated special:
	  (mapcar 
	   #'(lambda (entry)
	       (let* ((new-col (matrix-entry-col entry))
		      (new-col-row (assoc new-col
					  cone-n-cells
					  :test test)))
		 (assert new-col-row
			 ()
			 "No tagged entry for ~S."  new-col)
		 (make-matrix-entry (cdr new-col-row)
				    (car new-col-row)
				    (matrix-entry-value entry))))
	   closure-n-1-boundary))

    (setf paste-boundaries ; 0 <- 1 ... n-2 <- n-1
	  (nconc (mapcar 
		  #'(lambda (mappings)
		      (mapcar #'(lambda (mapping)
				  (make-matrix-entry (cdr mapping)
						     (car mapping)
						     1))
			      mappings))
		  (butlast (cdr cone-cellss))) ; -1 <- 0 is not wanted, last gets special treatment.
		 (list paste-n-1-boundary)))

    (pop closure-boundaries)
    (pop closure-cellss)

    (values
     (mapcar #'(lambda (cl-boundary paste-boundary cone-boundary)
		 (append cl-boundary paste-boundary
			 (chain* -1 cone-boundary)))
	     (append closure-boundaries '(()))
	     paste-boundaries
	     cone-boundaries)

     (mapcar #'(lambda (cells cone-cells)
		 (append  
		  cells
		  (map-into cone-cells #'cdr cone-cells)))
	     (nconc closure-cellss (list ()))
	     cone-cellss))))


'(define-symbol-macro test
 (triangulate-cell-2 '|ab|
		     '((((|ab| |a|) . -1) ((|ab| |b|) . 1)
			((|cd| |c|) . -1) ((|cd| |d|) . 1)
			((|ac| |a|) . -1) ((|ac| |c|) . 1)
			((|bd| |b|) . -1) ((|bd| |d|) . 1)))
		     '((|a| |b| |c| |d|) (|ab| |cd| |ac| |bd|))
		     :APEX '#:|X|)
)



(defun collapse-edge-star (vertex edge boundaries cellss &key (test #'eql))

  "Retracts the specified edge to the specified vertex (which must be close to that edge) 
and the closure of the edge neighbourhood to that neighbourhoods boundary part which is 
\"at the side\" if the specified vertex.

The other vertex and the corresponding \"side\" of that vertex will be removed and all 
\"outside\"-cells will be pasted onto the retraction. 

The cells in the edge-neighbourhod should be simplexes
This operation (still) is non-destructive."

  (declare (optimize (debug 3)))

  (let* ((old-boundaries (copy-tree boundaries)) ; to error-test: This function is 
	 (old-cellss     (copy-tree cellss))      ;   supposed to be non-destructive.

	 ;; a sequence of empty sets to be used to be used to pad other set sequences.
	 ;; It must be long enough to pad them all.
	 (no-cellss                    (mapcar (constantly ()) cellss))
	 (edge-bd                      (matrix-get-row (car boundaries) edge))

	 (#:x (print `(edge = ,edge)))
	 (#:x (print `(edge-bd = ,edge-bd)))

	 ;; the vertex which will collapse to the specified vertex.
	 ;;  This should be a singleton list (dropee).
	 (dropees                      (remove vertex edge-bd
					       :test  test
					       :key  #'matrix-entry-col))
	 (#:x (print `(dropees = ,dropees)))

	 ;; the vertex that will disappear after collapsing the edge.
	 (dropee                       (when dropees                            ; the vertex that will be removed after
					 (matrix-entry-col (car dropees))))     ;   the edge has collapsed.
	 (#:x (print `(dropee = ,dropee)))

	 ;; the star neighbourhood of the specified edge:
	 ;; these cells will be collapsed to zero and then removed.
	 edge-star-cellss              ; bound in following multitple-value-bind form.
	 (edge-star-boundaries         (multiple-value-bind (bds cls)
					   (get-star (list edge) (cdr boundaries) :test test)
					 (setf edge-star-cellss (cons () cls))
					 (cons () bds)))

	 ;; the closure of that star neighbourhood:
	 edge-star-closure-cellss     ; bound in ...
	 (edge-star-closure-boundaries (multiple-value-bind (bds cls)
					   (get-closure (car (last edge-star-cellss))
							(subseq boundaries 0 (length edge-star-boundaries))
							:test test)
					 (setf edge-star-closure-cellss cls)
					 bds))

	 (#:x (print `(edge-star-closure = (cellss -> ,edge-star-closure-cellss 
					    boundaries -> ,edge-star-closure-boundaries))))

	 

	 ;; the topological boundary of that star neighbourhood = closure minus star:
	 ;; It is called "surface" because it is the /topological/ boundary 
	 ;; not be confounded it with an /algebraic/ boundary operator.
	 (minus                        (rcurry #'set-difference :test test)) ; theoretically, :test eq should do, too
	 (bd-minus                     (rcurry #'set-difference 
					       :test 
					       #'(lambda(a-entry b-cell)
						   ;; TODO: Is the use of or correct?
						   (or (funcall test (matrix-entry-row a-entry) b-cell)
						       (funcall test (matrix-entry-col a-entry) b-cell))))) 
	 (edge-star-surface-cellss     (mapcar minus
					       edge-star-closure-cellss    ; closure minus
					       edge-star-cellss))          ; .. interior
	 (edge-star-surface-boundaries (mapcar bd-minus
					       edge-star-closure-boundaries
					       (mapcar #'append
						       (cdr edge-star-cellss) ; the row indexes
						       edge-star-cellss       ; the col indexes
						       )))

	 ;; the star of the dropee vertex within the star closure:
	 ;; This are the cells which will disappear when the neighbourhood collapses.
	 drop-side-cellss              ; bound in ...
	 (drop-side-boundaries         (multiple-value-bind (bds cls)
					   (get-star (list dropee)
						     edge-star-surface-boundaries :test test)
					 (setf drop-side-cellss cls)
					 bds))

	 ;; compute a list of replacements:
	 ;; a replacement is an matrix where the row indexes are the cells where each will be replaced 
	 ;; by the corresponding row.
	 (replacements    (mapcar #'(lambda (edge-star-closure-boundary 
					     edge-star-int-cells
					     drop-side-bd-cells
					     &aux (result ())) ; an empty matrix
				      (dolist (collapsing-cell edge-star-int-cells result)
					;; collapsing cell is the cell which is set to zero.
					(let* (;; then its boundary  must be zero, too:
					       (cell-boundary (matrix-get-row edge-star-closure-boundary
									      collapsing-cell))
					       ;; the boundary entries which are to be replaced:
					       (drop-boundary (intersection
							       (mapcar #'matrix-entry-col cell-boundary)
							       drop-side-bd-cells
							       :test test)))
					  ;; hence drop-boundary = drop-boundary - cell-boundary, as cell-boundary = 0
					  ;;                     = - (cell-boundary - drop-boundary)
					  (when drop-boundary 
					    ;; cell-boundary contains the corresponding matrix-entries:
					    (dolist (cell-incidence  cell-boundary)
					      (let ((boundary-cell (matrix-entry-col cell-incidence))
						    (-boundary-side (- (matrix-entry-value cell-incidence))))
						;; the subtraction "- drop-boundary" is done by just ignoring 
						;;   drop-side-bd-cells:
						(unless (member boundary-cell
								drop-side-bd-cells
								:test test)
						  (setf result
							(chain+ result
								(mapcar #'(lambda (drop-cell) 
									    (make-matrix-entry drop-cell
											       boundary-cell
											       -boundary-side))
									drop-boundary))))))))))
				  edge-star-closure-boundaries ; starts with edge-vertex
				  (cdr edge-star-cellss) ; start with an edge
				  drop-side-cellss       ; starts with a vertex
				  )))


    (let ((*print-escape* nil) (*print-readably* nil))

      (write-boundary edge-bd)
      (format t "~&Will drop ~W." dropee)

      (format t "~&The edge star is:")
      (mapcar #'(lambda (s) (terpri) (write-set s))
	      edge-star-cellss)
      (let ((dim 0))
	(mapcar (utils:mc-rcurry #'write-boundary :dimension (incf dim))
		edge-star-boundaries))

      (format t "~&The edge star closure is:")
      (mapcar #'(lambda (s) (terpri) (write-set s))
	      edge-star-closure-cellss)
      (let ((dim -1))
	(mapcar (utils:mc-rcurry #'write-boundary :dimension (incf dim))
		edge-star-closure-boundaries))

      (format t "~&The edge star surface is:")
      (mapcar #'(lambda (s) (terpri) (write-set s))
	      edge-star-surface-cellss)
      (let ((dim -1))
	(mapcar (utils:mc-rcurry #'write-boundary :dimension (incf dim))
		edge-star-surface-boundaries))

      (format t "~&The edge star surface's drop side is:")
      (mapcar #'(lambda (s) (terpri) (write-set s))
	      drop-side-cellss)
      (let ((dim -1))
	(mapcar (utils:mc-rcurry #'write-boundary :dimension (incf dim))
		drop-side-boundaries))

      (format t "~&Replacement:")
      (let ((dim -1))
	(mapcar (utils:mc-rcurry #'write-boundary :dimension (incf dim))
		replacements))

      (terpri))
      

    (assert dropees
	    (dropees)
	    "There must be a cell to remove.")

    (assert (null (cdr dropees))
	    (dropees)
	    "The cell to drop must be unique.")
    

    ;; now each edge-star-cell C is formally set to zero.
    ;; Then its boundary d(C) is zero, too.
    ;; but when a drop-side-cell x is in d(C), we have d(C) = 0 = (C:a)*a + (C:b)*b  + ... + (C:x)* x
    ;; and so have x = - 1/(C:x) * ( (C:a)*a + (C:b)*b + ...),
    ;; where {a, b, ...} are non-dropees.
    ;; So we replace every boundary Entry (N:x) by (N:a) + (N:b) + ...
    ;; Where (N:a) := -(N:x)*(C:a)/(C:x)
    ;; Where (N:b) := -(N:x)*(C:b)/(C:x)

    (let ((dim 0)(*print-escape* nil) (*print-readably* nil))
      (declare (ignorable dim))

      (setf edge-star-cellss (nconc edge-star-cellss no-cellss)
	    drop-side-cellss (nconc drop-side-cellss no-cellss)
	    replacements     (nconc replacements     no-cellss))

      (multiple-value-prog1

	  (values
	   ;; the result boundaries
	   (mapcar #'(lambda (boundary replacement-matrix row-collapsees col-collapsees row-dropees col-dropees)  
			(mapcan #'(lambda (old-entry) 
				    (let ((cell  (matrix-entry-row old-entry)))
				      (unless (or (member cell row-dropees :test test)
						  (member cell row-collapsees :test test))
					(let* ((bd-cell (matrix-entry-col old-entry))
					       (side    (matrix-entry-value old-entry))
					       (replace (matrix-get-row replacement-matrix 
									bd-cell 
									:test test)))
					  (assert (not (intersectionp 
							replace
							col-dropees
							:test #'(lambda (m-entry drop-cell)
								  (funcall test 
									   (matrix-entry-col m-entry)
									   drop-cell))))
						  ()
						  "Cell replacement must conatain collapsed cells.")
					  
					  (if replace
					      (progn (assert (and side (not (zerop side))))
						     ;; replace entry by the non-droppees
						      (mapcan #'(lambda (new-entry)
								   (let ((col (matrix-entry-col new-entry))
									 (val (matrix-entry-value new-entry)))
								     (unless (or (member col col-collapsees :test test)
										 (member col col-dropees    :test test))
								       (list (make-matrix-entry cell col (/ val side))))))
							      replace))
					      (list old-entry))))))
				boundary))
		    boundaries
		    replacements
		    (cdr edge-star-cellss)
		    edge-star-cellss
		    (cdr drop-side-cellss)
		    drop-side-cellss)
	   
	   (mapcar (reduced minus) cellss
		   edge-star-cellss
		   drop-side-cellss ))
	   
       (assert (equal old-boundaries boundaries)
	       ()
	       "Boundaries have been destroyed.")
       (assert (equal old-cellss cellss)
	       ()
	       "Cells have been destroyed.")))))


(define-symbol-macro test
 (collapse-edge-star '|a|
		     '(APEX |a|)
		     '(((((APEX |a|) |a|) . -1)
			(((APEX |b|) |b|) . 1)
			(((APEX |b|) APEX) . -1)
			(((APEX |a|) APEX) . 1)))
		     '((|b| |a| APEX) ((APEX |b|) (APEX |a|)))
		     :TEST #'equal))


(defun partitioning (lst &key from-end (test #'eql) (key #'identity))
  (let* ((result ()))
    (if from-end
	(dolist (item (reverse lst) result)
	  (let ((items (assoc (funcall key item) result
			      :test test 
			      :key key)))
	    (if items
		(nconc items `(,item))
		(push `(,item) result))))
	(dolist (item lst (nreverse result))
	  (let ((items (assoc (funcall key item)  result 
			      :test test
			      :key key)))
	    (if items
		(nconc items `(,item))
		(push `(,item) result)))))))


(defun collapse-test (test classes a b)

  "Returns true if both objects a and b either satisfy the test or if the first 
list that conatins object a also contains object b with respect to that specified test.

If nil is specified as test #'eql is used as a default.

This is an equivalence relation if (1) the specified test is an equivalence relation 
and (2) the specified classes are either disjoint or equal as sets with respect to that 
given test.

This function is primarily intented for currying into a two-argument-test. Therefore 
test and classes have been arranged as first arguments:

Example:
\(curry #'collapse-test #'eq '((a b c) (x y))\) returns a predicate p which is same as #'eq 
except for the symbols a, b, and c and for x and y which satisfy:
\(funcall p a b\) => true
\(funcall p a x\) => false
\(funcall p x y\) => true."

  (let ((test (if test test #'eql)))
    (or (funcall test a b)
	(member b 
		(find a classes 
		      :test #'(lambda (item set) 
				(member item set :test test)))
		:test test))))

(defun refine-test (finer-test coarser-test fine-objects a b)

  "Dual to collapse-test: Applies the finer test to a selection of objects,
otherwise the coarser test.
 
Returns true if a and b satisfy the finer test or if neither of them is member of the 
specified \"fine objects\" (with respect to the finer test) and satisfy the coarser test.

This is an equivalence relation if finer-test and coarser-test are equivalence relations 
and if finer-test is finer (stronger) than coarser-test:
 \(if (funcall finer-tests a b) (funcall coarser-test a b) T \)
must always be true."

  (or (funcall finer-test a b)
      (and (not (member a fine-objects :test finer-test))
	   (not (member b fine-objects :test finer-test))
	   (funcall coarser-test a b))))


(defun has-duplicates (lst &key (test #'eql) (from-end nil) test-not (start 0) (end nil) key)

  "True, if in lst exists a duplicate. The boolean value is equivalent to 
\(/= (length lst) (length (remove-duplicates lst ...))\).

All parameters are the same as in remove-duplicates and they have the same semantics."

  (if test-not
      (has-duplicates lst 
		      :test (complement test-not)
		      :key key
		      :from-end from-end
		      :start start :end end)
      (progn 
	(remove-duplicates lst 
			   :test #'(lambda (a b) 
				     (when (funcall test a b)
				       (return-from has-duplicates T)))
			   :key key
			   :from-end from-end
			   :start start :end end)
	nil)))


(defun collapse-duplicate-cells (boundaries cellss 
				 &key 
				 from-end 
				 (fine-test #'eql) 
				 (coarse-test fine-test) 
				 (sign (constantly 1)))

  "Removes duplicate cells in each cell set in cellss with respect to coarse-test by collapsing 
them into one cell. 

In each cell set of cellss the last occurrence of two cells satisfying the test will be kept in the 
result unless from-end is specified and true in which case the first occurrence of the cells will be 
kept in the result.

The boundary d will be adjusted accordingly:
For two duplicate cells a and a' where a' will be removed each occurence of a' in some boundary 
  d(A) = ... + xa'*a' + ... 
the summand xa'*a' will be replaced by *xa'*sign(a',a)*a using the specified sign function. 

The result may or may not be a complex. It is a complex when sign(a,a')*d(a) = d(a') up to 
the coarser equivalence."
  
  (let* ((dropeess          ())
	 (collapseess       ())
	 (result-classess   (mapcar (rcurry #'partitioning 
					    :from-end from-end 
					    :test coarse-test)
				    cellss))
	 (have-collapsed    (mapcar (curry #'some #'cdr) result-classess)))

    ;; this constructs the matrices which modifiy the boundaries. 
    (setf dropeess ; drop rows
	  (mapcar #'(lambda (dropees)
		      (mapcar #'(lambda (class &aux (row (car class)))
				  (make-matrix-entry row row 1))
			      dropees))
		  result-classess)

          collapseess ; identify columns
	  (mapcar #'(lambda (dropees)
		      (mapcan #'(lambda (class &aux (col (car class)) (row col))
				    (cons 
				     (make-matrix-entry row col 1)
				     (mapcar #'(lambda (row)
						 (make-matrix-entry row col
								    (funcall sign row col)))
					     (cdr class))))
			      dropees))
		  result-classess))

    
    (values 
     (mapcar #'(lambda (boundary dropees collapsees need-dropees need-collapsees)
		 (let ((result boundary))
		   (when need-dropees
		     (setf result (mult-boundary dropees result     :test fine-test)))
		   (when need-collapsees
		     (setf result (mult-boundary result  collapsees :test fine-test)))
		   result))
	     boundaries
	     (cdr dropeess)
	     collapseess
	     (cdr have-collapsed)
	      have-collapsed)
	     
     (mapcar #'(lambda (has-collapsed original-cells collapsed-cells)
		 (if has-collapsed
		     (mapcar #'car collapsed-cells)
		     original-cells))
	     have-collapsed
	     cellss
	     result-classess))))




;;; a test case for the above:
(defparameter *collapse-squares-edge-cells* 
;; A ---AB---> B
;; |     -     |
;;AC+  ABCD  - BD
;; v     +     v
;; C ---CD---> D
;; C<---DC---  D
;; |     +     |
;;CE+  CDEF  - DF
;; v     +     v
;; E ---EF---> F

  '((A B C D E F)
    (AB AC BD CD DC CE DF EF)
    (ABCD CDEF)))

(defparameter *collapse-squares-edge-boundary*
  '(
    ( ((AB A) . -1) ((AB B) . +1)
      ((AC A) . -1) ((AC C) . +1)
      ((BD B) . -1) ((BD D) . +1)
      ((CD C) . -1) ((CD D) . +1)
      ((DC C) . +1) ((DC D) . -1)

      ((CE C) . -1) ((CE E) . +1)
      ((DF D) . -1) ((DF F) . +1)
      ((EF E) . -1) ((EF F) . +1)
     )
    (((ABCD AB) . -1)
     ((ABCD AC) . +1)
     ((ABCD BD) . -1)
     ((ABCD CD) . +1)

     ((CDEF DC) . +1)
     ((CDEF CE) . +1)
     ((CDEF DF) . -1)
     ((CDEF EF) . +1)
     )
    ))


(print 
 (collapse-duplicate-cells *collapse-squares-edge-boundary*
			   *collapse-squares-edge-cells*
			   :fine-test #'eq
			   :coarse-test (curry #'collapse-test #'eq '((CD DC)))
			   :sign #'(lambda (a b) 
				     (cond ((and (eq a 'cd) (eq b 'dc)) -1)
					   ((and (eq a 'dc) (eq b 'cd)) -1)
					   (T +1)))))
			  


(defparameter *triangulate-square-points* 
;; A ---AB---> B
;; |     -     |
;;AC+  ABCD  - BD
;; v     +     v
;; C ---CD---> D
;; |     -     |
;;CE+  CDEF  - DF
;; v     +     v
;; E ---EF---> F

  '((A B C D E F)
    (AB AC BD CD CE DF EF)
    (ABCD CDEF)))

(defparameter *triangulate-square-boundary*
  '(
    ( ((AB A) . -1) ((AB B) . +1)
      ((AC A) . -1) ((AC C) . +1)
      ((BD B) . -1) ((BD D) . +1)
      ((CD C) . -1) ((CD D) . +1)

      ((CE C) . -1) ((CE E) . +1)
      ((DF D) . -1) ((DF F) . +1)
      ((EF E) . -1) ((EF F) . +1)
     )
    (((ABCD AB) . -1)
     ((ABCD AC) . +1)
     ((ABCD BD) . -1)
     ((ABCD CD) . +1)

     ((CDEF CD) . -1)
     ((CDEF CE) . +1)
     ((CDEF DF) . -1)
     ((CDEF EF) . +1)
     )
    ))

(print (boundaryp *triangulate-square-boundary*))

(mapc #'print (multiple-value-list (get-closure '(ABCD) *triangulate-square-boundary*)))


(define-symbol-macro test 
    (print (triangulate-cell 'p 
			 'ABCD 
			  *triangulate-square-boundary*
			  *triangulate-square-points*
			  :make-tensor #'cons
			  :tensor-pi-2 #'cdr
) ))

(define-symbol-macro test-2
    (print (triangulate-cell-2  
			 'ABCD 
			  *triangulate-square-boundary*
			  *triangulate-square-points*
			  :apex 'p) ))

(print (triangulate-cell 'apex 
			 'ABCD 
			 *triangulate-square-boundary*
			 *triangulate-square-points*) )

(print (boundaryp (triangulate-cell 'A 
			 'ABCD 
			  *triangulate-square-boundary*
			  *triangulate-square-points*) ))

(collapse-edge-star 'A 'AB *triangulate-square-boundary* *triangulate-square-points*)
;; => values
'((((AC A) . -1) ((AC C) . 1) 
   ((CD C) . -1) ((CD D) . 1) 
   ((CE C) . -1) ((CE E) . 1)
   ((DF D) . -1) ((DF F) . 1) 
   ((EF E) . -1) ((EF F) . 1))
  (((CDEF CD) . -1) ((CDEF CE) . 1) ((CDEF DF) . -1) ((CDEF EF) . 1)))
;; and
'((F E D C A) (AC CD CE DF EF) (CDEF))


(collapse-edge-star 'A 'AC *triangulate-square-boundary* *triangulate-square-points*)
;; =>
'((((AB A) . -1) ((AB B) . 1) 
   ((BD B) . -1) ((BD D) . 1) 
   ((CE A) . -1) ((CE E) . 1)
   ((DF D) . -1) ((DF F) . 1) 
   ((EF E) . -1) ((EF F) . 1))
  (((CDEF AB) . -1) ((CDEF BD) . -1) ((CDEF CE) . 1) ((CDEF DF) . -1) ((CDEF EF) . 1)))

'((F E D B A) (AB BD CE DF EF) (CDEF))

(defparameter *1-sphere-cells*
  '((vertex) (edge)))

(defparameter *1-sphere-boundary*
  '(;;Sequence of matrices 
     (;; a Matrix
      ((edge vertex) . 0) ; matrix entry
)))

(triangulate-cell 'apex
		  'edge 
		  *1-sphere-boundary*
		  *1-sphere-cells*)

;;; http://www.neglook.com/movies/manual-intro.mov a newLISP-commercial.
 
(defun complex-triangulation (boundary cells &key (test #'eql) (verbose T)
			      &aux ;(list-test (curry #'list-test test))
			           (tree-test (curry #'tree-test test)))

  "Algebraically triangulates a complex. 
Cells and boundary are ordered from dimension 0 to dimension n."

  (do* ((i 1 (1+ i)) ; i is dimension. May be removed later.
	(0-cells         (car cells))
	(i-cells-iter    (cdr cells)              (cdr i-cells-iter))
	;; the current i-cells (input)
	(i-cells         (car  i-cells-iter)      (car  i-cells-iter))

	(0-boundary      (car boundary))
	(i-bd-iter       (cdr boundary)           (cdr i-bd-iter))
	;; the current i-boundary: i+1 --> i (input)
	(i-boundary      (car i-bd-iter)          (car i-bd-iter))
	(i-boundary*     (matrix-copy i-boundary) (matrix-copy i-boundary))     ;

	;; the previous i-boundary: i --> i-1 (input)
	(i-1-bd-iter     boundary                 (cdr i-1-bd-iter))
	(i-1-boundary    0-boundary               (car i-1-bd-iter))

	;; the map (er relation) back from the simplexes to the original cells:
	(i-1-cell-map    (mapcar #'(lambda (0-cell)           
				     (make-matrix-entry 
				      (list 0-cell)           ; row:   The new 0-simplex ...
				      0-cell                  ; col:   ... mapped to its old 0-cell
				      0))                     ; value: ... of dimension 0
				 0-cells))

	(result-cell-maps (list i-1-cell-map))
	(new-i-cells->vertices ())                              ; will be updated in the body
	;; the current map from the i-cells to the vertices used for triangulation.
	(i-cells->vertices (mapcar #'(lambda (entry)
				       (let ((vtx (matrix-entry-col entry)))
					 (cons vtx vtx)))       ; the vertex maps to itself
				   i-1-cell-map)
			   (shiftf new-i-cells->vertices ()))

	;; the following is the current boundary from the i+1-cells to the i-simplexes.
	;;   its row indexes are original cells, its column indexes are tensors:
	(i-1-boundary*     (mapcar #'(lambda (entry)                       ; the i-th boundary from the remaining 
				       (make-matrix-entry                  ;   complex cells to the new simplices.
					(matrix-entry-row entry)           ; copy the i+1 row entry
					(matrix-entry-row                  ; use the new i-cell ...
					 (car (matrix-get-col i-1-cell-map ; .. from the cell-map
							      (matrix-entry-col entry))))
					(matrix-entry-value entry)))
				   i-1-boundary)
			   #|will be updated in the body|#)
	(result-co-boundaries ())  ; this is where the co-boundaries are collected. starting from 1-2
	(result-co-cells      (mapcar #'matrix-entry-row  i-1-cell-map))  ; this is where the co-cells are collected.


	);end of do*-list

       ((progn
	  ;; begin space for debug stuff

	  (when verbose 
	    (format t "~%~S-cells = ~A~%" i i-cells)
	    (let ((*print-escape*  nil) (*print-readably* nil))
	      (mapc (rcurry #'write-boundary :symbol "res-d") result-co-boundaries))
	    )


	  ;; end space for debug stuff
	(null i-cells-iter)) ;; TODO: should possibly be (null i-bd-iter)?
	(values (nreverse result-co-boundaries)
		(nreverse result-co-cells)))


    (push () result-co-boundaries) ; increment the result dimension by one
    (push () result-co-cells)      ; increment the result dimension by one

    ;; triangulate each i+1-cell:
    (dolist (i-cell i-cells)
      (let (bd* bd-cell  co-bd*  co-bd-cells*  cell-vertex  vertex i-simplex-co-boundary  new-co-simplexes  i-star)
	
	;; the closure of the i-cell will be comuted in two steps:
	;; Step 1: find the (algebraic) i+1-cell-boundary (using test)
	(setf bd* (matrix-get-row  i-1-boundary*
				   i-cell
				   :test test))
	;; Step 2: find the above's boundaries co-closure within the triangulation (using tree-test)
	(multiple-value-setq (co-bd* co-bd-cells*)
	  (get-co-closure (delete-duplicates (mapcar #'matrix-entry-col bd*)
					     :test tree-test)
			  (cdr result-co-boundaries) ; forget the trailing ()
			  :test tree-test))

	;; now find the triangulation apex: cell-vertex:
	(setf bd-cell nil ; (this is for the slime formatter)
	      ;; chose one matrix entry in the (original) boundary if i-cell
	      bd-cell     (matrix-entry-col (find i-cell i-1-boundary 
						  :test test
						  :key  #'matrix-entry-row))
	      ;; choose one boundary element and ...                    
	      cell-vertex (assoc bd-cell i-cells->vertices :test test) 
	      ;; ...take its vertex.
	      vertex      (cdr cell-vertex))
	(unless cell-vertex
	  (warn "There was no vertex for ~S" i-cell)
	  (setf vertex i-cell))
	;; and register it for the next iteration:
	(setf new-i-cells->vertices 
	      (acons i-cell vertex new-i-cells->vertices))

	;; now the actual triangulation of i-cell takes place:
	(when verbose
	  (print `(vertex = ,vertex & i-cell = ,i-cell))
	  (let ((*print-readably* nil) (*print-escape* nil))
	    (mapc #'write-boundary (cons bd* co-bd*)))
	  (terpri))

	(assert (co-boundaryp (cons bd* co-bd*) :test tree-test :before 'entry)
		(bd* co-bd*)
		"co-boundaryp on entry")
	(multiple-value-setq (i-simplex-co-boundary new-co-simplexes)
	  (co-triangulate-cell vertex 
			       i-cell 
			       (cons bd* co-bd*)
			       `((,i-cell) ,@co-bd-cells*) ; taken from boundaries. This must be changed later.
			       :test tree-test))
	(assert (co-boundaryp i-simplex-co-boundary :test tree-test :after 'exit)
		(i-simplex-co-boundary)
		"co-boundaryp on exit")
	
	;; now add the triangulation to the result:
	(setf result-co-boundaries (map-into result-co-boundaries
					     (rcurry #'chain+ :test tree-test)
					     result-co-boundaries
					     i-simplex-co-boundary)
	      result-co-cells      (map-into result-co-cells
					     #'(lambda (&rest args)
						 (reduce (rcurry #'union :test tree-test) 
							 args))
					     result-co-cells
					     new-co-simplexes
					     (cons () co-bd-cells*)))
	;; now modify the forthcoming boundary:
	(setf i-star (shiftf (matrix-get-col i-boundary* i-cell :test test) ()))
	(dolist (new-simplex (car new-co-simplexes))
	  (setf (matrix-get-col i-boundary* new-simplex :test test)
		i-star))))
      
    ;; now the modified star must become the new boundary in the next iteration:
    (setf i-1-boundary* i-boundary*)
    

    );end of do*
  )


;;; Some test cases:
;;;
;;; A vertex /a/.
;;;
(defparameter *sample-vertex-cells*    
  '((\aq)))
(defparameter *sample-vertex-boundary* 
  '() )

(progn ; The vertex test case
  (mapcar #'printc 
	  (mapcar #'list 
		  '(|bd:| |cells:| |map:|)
		  (multiple-value-list 
		   (complex-triangulation *sample-vertex-boundary*
					  *sample-vertex-cells* ))))
  (values))

;;;
;;;      ab
;;; a ---------> b
;;;
(defparameter *sample-edge-cells*    
  '((\a \b) (|ab|)))
(defparameter *sample-edge-boundary* 
  '(( ((|ab| \a) . -1) 
      ((|ab| \b) . +1)) ))

(progn ; The edge test case
  (mapcar #'printc 
	  (mapcar #'list 
		  '(|bd:| |cells:| |map:|)
		  (multiple-value-list
		   (complex-triangulation 
		    *sample-edge-boundary*
		    *sample-edge-cells* 
		    ))))
  (values))


;;;  a ----> b
;;;  |   -   |
;;;  |+  A  -|
;;;  V   +   V
;;;  c ----> d


(defparameter *sample-square-cells* 
  '((\a \b \c \d)
    (|ab| |cd| |ac| |bd|)
    (A)))

(defparameter *sample-square-boundary*
  '((((|ab| \a) . -1) ((|ab| \b) . +1)
     ((|cd| \c) . -1) ((|cd| \d) . +1)
     ((|ac| \a) . -1) ((|ac| \c) . +1)
     ((|bd| \b) . -1) ((|bd| \d) . +1))
    (((A |ab|) . -1)
     ((A |ac|) . +1)
     ((A |cd|) . +1)
     ((A |bd|) . -1) )))


(defun square-test-case-2 ()
  (let* ; The square test case
    ((cells         (copy-tree *sample-square-cells*))
     (boundary      (copy-tree *sample-square-boundary*))
     (triangulation (multiple-value-list (complex-triangulation boundary cells)))
     (new-boundary  (car triangulation))
     (boundaryp     (boundaryp new-boundary :test #'equal)))

  (mapcar #'printc 
	  (mapcar #'list 
		  '(|bd:| |cells:| |map:|)
		  triangulation))

  (printc `(boundary-p = ,boundaryp))

  (unless boundaryp
    (format t "~%bd^2 = ~A "
	    (mapcar #'(lambda (a b) 
			(remove-if #'zerop (mult-boundary a b :test #'equal)
				   :key #'cdr))
		    (cdr new-boundary)
		    new-boundary)))

  '(assert (equal cells    *sample-square-cells*))
  (unless (equal cells *sample-square-cells*)
    (format T "~%cells are ~A. " cells)
    (format T "~%but they were ~A. " *sample-square-cells*))
  (assert (equal boundary *sample-square-boundary*))

  (values)))
 
'(square-test-case-2)


'(let ; The above cube test-case:
    ((cells     (copy-tree *sample-cube-cells*))
     (boundary  (copy-tree *sample-cube-boundary*)))

  (mapcar #'printc 
	  (mapcar #'list 
		  '(|cells:| |bd:| |map:|)
		  (complex-triangulation cells boundary)))
  (assert (equal cells    *sample-cube-cells*))
  (assert (equal boundary *sample-cube-boundary*))
  (values))


(defun complexes-union (boundaries-1 cellss-1 boundaries-2 cellss-2 &key (test #'eql))
  (let* ((dimension (max (length boundaries-1) (length boundaries-2)))
	 (result-cellss     (make-list (1+ dimension)))
	 (result-boundaries (make-list dimension)))
    (setf result-cellss     (map-into result-cellss #'identity cellss-1)
	  ;; TODO: Append all cellss-1 to the /beginning/ of cellss-2.
	  result-cellss     (map-into result-cellss (rcurry #'union :test test)  
				      result-cellss cellss-2)
	  result-boundaries (map-into result-boundaries #'identity boundaries-1)
	  result-boundaries (map-into result-boundaries 
				      (rcurry #'chain+ :test (matrix-key-test test))
				      result-boundaries
				      boundaries-2))
    (unless (boundaryp result-boundaries :test test)
      (error "Cannot unify these complexes."))
    
    (values result-boundaries
	    result-cellss)))



(defun new-complex-triangulation (boundaries cellss &key (test #'eql))
  "An algebraic triangulation of the complex given by boundaries and cellss."

  ;; Strategy: 
  ;; A triangulation is the replacement of cells by lists of vertices
  ;; and a map (morphism) back from these triangles to the original cells.
  ;; Iteration:
  ;;   So each vertex x will be replaced by a list (x).
  ;;   Each edge e with boundary d(e)=t - s will be replaced by (s t).
  ;;   Each cell x will be triangulated by first using an arbitrary apex.
  ;;     (via triangulate-cell)
  ;;   Then the apex will be identified with one boundary vertex.
  ;;     (via collapse-edge-star)
  ;;   Then duplicate simplexes will be collapsed 
  ;;     (via collapse-duplicate-cells)
  ;;   Then the new n-simplexes will be pasted onto the n+1-cells.
  ;;   The algorithm then iterates from dimension i to i+1.

  (do* ((apex                '#:apex)
	;; the apex never satisfies the test with another object, except with itself.
	(apex-test           #'(lambda (a b)
				 (or (eq a b)
				     (and (not (eq a apex)) (not (eq b apex)) 
					  (funcall test a b)))))
	(tensor-test         (curry #'tensor-test  apex-test))
	(simplex-test        (curry #'simplex-test test)) ; apex is not used for simplex tests.
	(simplex-sign        (rcurry #'permutation-sign :test test))

	(vertices            (mapcar #'list (car cellss)))
	(new-cell-vertex-map ()) ; updated during iteration
	(cell-vertex-map     (mapcar #'(lambda (vtx) (cons vtx (car vtx)))
				     vertices)
			     (shiftf new-cell-vertex-map ())) ;; read and set to nil.
	(next-cellss         (cdr cellss)          (cdr next-cellss))
	(next-cells          (car next-cellss)     (car next-cellss))
	(next-boundaries     boundaries            (cdr next-boundaries))
	(next-boundary       (car next-boundaries) (car next-boundaries))

	(next-current-boundary ())
        (current-boundary    (mapcan #'(lambda (entry)
					 (format T "~&  entry: ~A" entry)
					 (let* ((col     (matrix-entry-col entry))
						(new-col (assoc col vertices :test #'eq)))
					   (when new-col
					     (list (make-matrix-entry 
						    (matrix-entry-row entry)
						    new-col
						    (matrix-entry-value entry))))))
				     next-boundary)
			     (shiftf next-current-boundary ()))

	;; the complex with all but the last module (cell set) triangulated:
	(current-boundaries (list current-boundary))
	(current-cellss     (list vertices next-cells))

	;; the fully triangulated complex:
	(result-boundaries ())
	(result-cellss     (list vertices)))

       ((null next-boundaries) (values result-boundaries result-cellss))

    ;; the triangulation complex consists of 
    ;;  the input complex C:
    ;;   C0 <--d0--- C1 <--d1--- ... Ci <--di--- ... <--dn-1-- Cn
    ;;                                           /
    ;;                                         /
    ;;  the                current-boundary  /                                  
    ;;                                     /
    ;;  and                             ./
    ;;   T0 <--t0--- T1 <--t1--- ... Ti '´
    ;;  the partially triangulated complex T.

    (dolist (cell next-cells)
      (format t "~&  cell: ~A" cell)
      (let* ((cell-bd     (matrix-get-row current-boundary cell :test test))
	     (cell-co-bd  (matrix-get-col next-boundary    cell :test test))
	     (bdcell      (matrix-entry-col (first cell-bd)))
	     (cell-vertex (assoc bdcell cell-vertex-map :test tensor-test))
	     (new-cell-boundaries ())
	     (new-cell-cellss     ()))

	(format t "~&  bdcell:~A" bdcell)
	(format t "~&  cell-vertex:~A" cell-vertex)

	;; Put a cone with given apex at boundary.
	(setf (values new-cell-boundaries new-cell-cellss)
	      (triangulate-cell-2 cell 
				  current-boundaries
				  current-cellss
				  :apex apex
				  :test tensor-test))

	(format t "~&  triangulation: ~A ~
                   ~&                 ~A" new-cell-cellss
		   new-cell-boundaries)

	;; find a vertex:
	(if cell-vertex
	    (setf cell-vertex (cdr cell-vertex))
	    (progn 
	      ;; TODO: replace by error after debugging:
	      (warn "There is no vertex for cell ~S~&(It will be used as its vertex)" cell)
	      (setf cell-vertex cell)))

	;; Now remove the apex.
	;; TODO: in case of the above warning/error the apex should not be collapsed but replaced.
	(setf (values new-cell-boundaries new-cell-cellss)
	      (collapse-edge-star cell-vertex 
				  `(,apex ,cell-vertex) 
				  new-cell-boundaries new-cell-cellss
				  :test tensor-test))
	;; Now we have a triangulated cell which can be pasted to the result.
	;; Note that pasting is a modification across all dimensions.

	;; Paste triangulated cell to the result complex:
	(setf (values result-boundaries result-cellss)
	      (complexes-union result-boundaries   result-cellss
			       new-cell-boundaries new-cell-cellss
			       :test tensor-test))

	;; Remove duplicate simplexes that only differ by permutation:
	;; TODO: Try to avoid this:
	(setf (values result-boundaries result-cellss)
	      (collapse-duplicate-cells 
	       result-boundaries result-cellss
	       :from-end     t
	       :fine-test    tensor-test
	       :coarse-test  simplex-test
	       :sign         simplex-sign))
	
	;; Now compute the next-current-boundary:
	(setf next-current-boundary 
	      (chain+ next-current-boundary
		      (mapcan #'(lambda (entry 
					 &aux 
					 (row (matrix-entry-row entry))
					 (value (matrix-entry-value entry)))
				  (mapcar #'(lambda (simplex)
					      (make-matrix-entry row simplex value))
					  (car (last new-cell-cellss))))
			      next-boundary
			      )
		      :test tensor-test))
	))

    );*od

)

;;; a test case for complex-triangulation:

(ignore-errors
  (print (new-complex-triangulation
	  *sample-square-boundary*
	  *sample-square-cells*)))


(defun tensor-string (ab &key (factor-formatter #'princ-to-string) (op "⊗"))
  (if ab
      (tensor-string-aux ab factor-formatter op)
      "()"))


(defun tensor-string-aux (ab factor-formatter op)
  (cond ((null ab) "")
	((atom ab) (funcall factor-formatter ab))
	((cdr ab)
	 (concatenate 'string 
		      (funcall factor-formatter (car ab))
		      op
		      (tensor-string-aux (cdr ab) factor-formatter op)))
	(t (funcall factor-formatter (car ab))))
)

(printc (tensor-string '(a b c . d)))

(defun chain-string (chain &key (cell-formatter #'princ-to-string) drop-zeros)
  (when drop-zeros
    (return-from chain-string 
      (chain-string (remove-if #'zerop chain :key #'cdr)
		    :cell-formatter cell-formatter
		    :drop-zeros     nil)))
  (when (null chain)
    (return-from chain-string "0"))
  (let ((is-minus (minusp (cdar chain)))
	(result ""))
    (dolist (item chain)
      (let* ((summand (funcall cell-formatter (car item)))
	     (factor  (abs (cdr item)))
	     (sign    (signum (cdr item)))
	     (op      (if (minusp sign) " - " " + ")))
	
	(setf result (concatenate 'string
				  result
				  op 
				  (if (eql factor 1) 
				      ""
				      (format nil "~S·" factor))
				  summand))))
    (if (not is-minus)
	(subseq result 3)
	result)))


(print (chain-string '((a . 1) (b . 4) (nix . 0))
		     :drop-zeros t))

(printc (chain-string (first *sample-edge-boundary*) :cell-formatter #'tensor-string))

(defun multiple-value-mapcar (fxn list &rest lists)
  (values-list 
   (%mv-map-transpose 
    (%mv-map-aux #'mapcar fxn (cons list lists)))))


(defun  multiple-value-maplist (fxn list &rest lists)
  (values-list 
   (%mv-map-transpose 
    (%mv-map-aux #'maplist fxn (cons list lists)))))


(defun %mv-map-aux (mapper fxn lists)
  (apply mapper 
	 #'(lambda (&rest args) 
	     (multiple-value-list (apply fxn args)))
	 lists))

(defun %mv-map-transpose (lists)
  (let ((result (mapcar #'list (car lists))))
    (dolist (in-lists (cdr lists) (mapcar #'nreverse result))
      (setf result (map-into  result  #'cons  in-lists result)))))

