;;;;  utils.lisp - Miscellaneous utilities.
;;;;
;;;;  Copyright (C) 2010  Norbert Paul
;;;;
;;;;  This program is free software: you can redistribute it and/or modify
;;;;  it under the terms of the GNU General Public License as published by
;;;;  the Free Software Foundation, either version 3 of the License, or
;;;;  (at your option) any later version.
;;;;
;;;;  This program is distributed in the hope that it will be useful,
;;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;  GNU General Public License for more details.
;;;;
;;;;  You should have received a copy of the GNU General Public License
;;;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;;;
;;;;  Contact: norbert.paul@kit.edu

(defpackage #:utils
  (:use :cl)
  (:export 
   #:traceexpr
   #:*DEBUG*
   #:set-equals 
   #:sorted-subsetp 
   #:sorted-set-equals
   #:equivalent
   #:boolean-equivalence
   #:lexicographic
   #:compare
   #:safe-sort
   #:safe-stable-sort
   #:flatten
   #:members
   #:members-if
   #:make-search-test
   #:assocs
   #:with-cons-places
   #:list-length<  #:list-length>  #:list-length=  #:list-length<= #:list-length>=
   #:compose
   #:curry #:rcurry
   #:mc-curry #:mc-rcurry
   #:supsetp
   #:flet-function-symbols
   #:remove-nth #:delete-nth
   #:defun-with-default-output-keys
   #:multiple-value-let
   #:project
   #:intersectionp
   #:reduced
   #:obsolete))

(in-package #:utils)

(defparameter *DEBUG* nil)

(defmacro traceexpr (xpr &optional (otherxpr xpr) (debug *DEBUG*))
  (if debug

      ;; print expr and result
      (let ((resultvar (gensym)))
	`(let ((,resultvar ,xpr)) 
	   '(format T "~%~S~%  ==> ~S " ',otherxpr ,resultvar)
	   ,resultvar))

      ;; identity
      xpr)
)

(defun make-test (key test &optional test-not)

" Generates a function f(a b) which executes either expression 
 (test (key a) (key b)) or (not (test-not (key a)(key b)))
 from the specified function designators key, test and test-not.

 An argument is said to be 'specified' if it has a non-nil value.

 It is an error to specify both test and test-not. 

 The use of test-not is deprecated and a warning will be issued if 
 it is specified.

 If an argument is not specified (by passing nil as its value) the 
 following defaults will be  used instead:
   key      defaults to #'cl:identity 
   test     defaults to #'cl:eql 
   test-not defaults to (complement #'cl:eql)."

 (let ((no-key-test 
	 (cond (test     test) 
	       (test-not (complement test-not)) 
	       (t        #'eql))))
     (if key
	 #'(lambda (a b)
	     (funcall no-key-test
		      (funcall key a)
		      (funcall key b)))
	 no-key-test)
     )
)


(defun make-2keys-test (key-1 key-2 test &optional test-not)

" Generates a function f(a b) which executes either expression 
 (test (key-1 a) (key-2 b)) or (not (test-not (key-1 a)(key-2 b)))
 from the specified function designators key-1, key-2, test, and 
 test-not.

 An argument is said to be 'specified' if it has a non-nil value.

 It is an error to specify both test and test-not. 

 The use of test-not is deprecated and a warning will be issued if 
 it is specified.

 If an argument is not specified (by passing nil as its value) the 
 following defaults will be  used instead:
   key-1 key-2 default  to #'cl:identity 
   test        defaults to #'cl:eql
   test-not    defaults to (complement #'cl:eql)."

 (let ((no-key-test 
	 (cond (test     test) 
	       (test-not (complement test-not)) 
	       (t        #'eql))))
     (if key-1
	 (if key-2 
	     #'(lambda (a b)
		 (funcall no-key-test
			  (funcall key-1 a)
			  (funcall key-2 b)))
	     #'(lambda (a b)
		 (funcall no-key-test
			  (funcall key-1 a)
			  b)))
	 (if key-2 
	     #'(lambda (a b)
		 (funcall no-key-test
			  a
			  (funcall key-2 b)))

	     no-key-test))
     )
)

(defun make-search-test (key test test-not)
  (make-2keys-test nil key test test-not))

(defun set-equals (list-1 list-2 &key key test test-not)

  (let ((effective-test (make-test key test test-not)))

     (and (subsetp list-1 list-2 :test effective-test)
	  (subsetp list-2 list-1 :test effective-test)))
)


(defun sorted-subsetp (list-1 list-2 &key key test test-not)

" Returns true if every element of list-1 matches some element of list-2 in the same order, 
 and false otherwise. Consecutive duplicates are handled as one element. 

 Formally this is a test if list-1 after removing consecutive equal elements 
 is a projection of list-2.

 The arguments  key, test, an test-not are used to specify equality according to 
 cl:subsetp
"

  (cond ((null list-1)   T)
	((null list-2) nil)
	(T (sorted-subsetp-aux list-1 list-2 (make-test key test test-not)))))


(defun sorted-subsetp-aux (list-1 list-2 test)

  (let ((lst2 list-2))
    (dolist (item1 list-1)
      (setf lst2 (member item1 lst2 :test test))
      (if (null lst2)
	  (return-from sorted-subsetp-aux nil)))
    T ))


(defun sorted-set-equals (list-1 list-2 &key key test test-not)
" Returns true if every element of list-1 matches some element of list-2 in the same order, 
 and vice versa. Consecutive duplicates are considered as one element.

 The arguments  key, test, an test-not are used to specify equality according to 
 cl:subsetp
"
  (cond((and (null list-1) (null list-2)) T)
       ((or (null list-1) (null list-2)) nil)
       (T
	(let ((test (make-test key test test-not)))
	  (and (sorted-subsetp-aux list-1 list-2 test)
	       (sorted-subsetp-aux list-2 list-1 test))))))

(defmacro equivalent (&body predicates)
  "Returns true if all args have the same boolean value."

  (cond ((null predicates) T)       ; if there is no predicate all are equivalent
	((null (cdr predicates)) T) ; one predicate is equivalent (to itself)
	(T 
	 `(if ,(car predicates)                   ; if   the first predicate is true
	       (and ,@(cdr predicates))           ; then they must all be true
	       (not (or ,@(cdr predicates)))))))  ; else they must all be false

(defun boolean-equivalence (value &rest values)
  "Returns true if all args have the same boolean value."

  (if value
      (every #'identity values)
      (every #'null     values)))


(defun lexicographic (func a b)
" Returns a function which defines the (strict) lexicographic order of 
 lists defines by the specified function func. That function nust be 
 a strict (irreflexive) linear order, hence, (funcall func x x) must 
 always return false, and if x /= y then 
 (or (funcall func x y) (funcall func y x)) must return true.

Examples:
 (funcall (lexicographic #'<) '(1 2 3) '(1 2)) 
==> true

 (funcall (lexicographic #'<) '(1 2 3) '(1 2)) 
==> nil
 (funcall (lexicographic #'<) '(1 2 3) '(1 3)) 
==> true
"
      (cond ((null a) b)
	    ((null b) nil)
	    (T 
	     (do* ((ra  a        (cdr ra))
		   (rb  b        (cdr rb))
		   (ra-null (null ra) (null ra))
		   (rb-null (null rb) (null rb))
		   (ai (first a) (first ra))
		   (bj (first b) (first rb)) )
		 ((or ra-null rb-null) (not rb-null))
	       (if (funcall func ai bj) ; if strictly ordered here
		   (return T)           ; a and b are ordered
		   (when (funcall func bj ai) ; else if inverse ordered
		     (return nil)))
	       ;; otherwise ai and bj are equal.
	       )))
)

(defun compare (a b &key (test #'<))
" Returns an integer value -1, 0, or 1 if a is less than, equal, or 
 greater than b, according to the specified test. 

 The semantics are exactly the same as Java's Comparator.compare(a, b), 
 or Comparable.compareTo(b) if the specified test denotes a linear order 
 (which, however, need neither be reflexive nor irreflexive)."
  (let ((a-b (if (funcall test a b) -1 0))
	(b-a (if (funcall test b a)  1 0)))
    (+ a-b b-a)) 
)

(defun safe-sort (seq pred &key key)
 (if (null key)
     (sort (copy-seq seq) pred)
     (sort (copy-seq seq) pred :key key))
)

(defun safe-stable-sort (seq pred &key key)
 (if (null key)
     (stable-sort (copy-seq seq) pred)
     (stable-sort (copy-seq seq) pred :key key))
)



(defun flatten (lst)
"The classic homework assignment."
  (cond ((null lst) lst)
	((atom lst) (list lst))
	(t (nconc (flatten (car lst)) (flatten (cdr lst))))))


(defun tail-flatten (lst &optional (result ()))
"The classic homework assignment -- tail-recursive."
  (cond ((null lst) result)
	((atom lst) (cons lst result))
	(t (tail-flatten (car lst) 
			 (tail-flatten (cdr lst) result)))))

(defun members (item list &key key test test-not)
  "Same as (remove item list :test (complement test)) etc.
 (was written before reading PCL :)."
  (let ((test (make-2keys-test nil key test test-not)))
    (mapcan #'(lambda (x) (if (funcall test item x)
			      (list x)
			      ()))
	    list )))

(defun members-if (predicate list &key key)
  "(see members :)."
  (let ((test (make-test key predicate)))
    (mapcan #'(lambda (x) (if (funcall test x)
			      (list x)
			      ()))
	    list ))
)

(defun assocs (item list &key key test test-not)
  "(: :)"
  (members item list 
	   :key #'car 
	   :test (make-test key test test-not))
)


(defun compose (&rest functions)
  "Creates a function (f1 (f2 (f3 x))) from a list of functions (f1 f2 f3)."
  (let ((functions (reverse (remove #'cl:identity (remove 'cl:identity functions)))))
    (if (null functions)
	#'cl:identity
	#'(lambda (x)
	    (let ((result x))
	      (dolist (fx functions result)
		(setf result (funcall fx result))))))))

;;; (with-cons-places ((willi otto (heinrich)) (() . otto-heirich))
;;;   form)
;;;
;;; should macroexpand in equivalent to
;;;
;;; (macrolet ((willi (x)     `(car ,x))
;;;            (otto  (x)     `(cadr ,x))
;;;            (heinrich (x)  `(caaddr ,x))
;;;            (otto-heinrich `(cdr ,x)))
;;;  . form )
	   

;; The initial macrolet version.
;; Shortcoming: There is no setf and the accessor macros cannot be passed as 
;;  function objects, say as :key arguments to searching functions.
;(defmacro with-tree-accessors (lst &rest forms)
;  (let* ((accessors (mapcan #'tree-accessors lst))
;	 (names     (mapcar #'car accessors))
;	 (macros (mapcar #'(lambda (name)
;			     (let ((accessor 
;				    (pack-tree-accessors 
;				     (cdr (assoc name accessors)))) ; gives first occurrence the precendence (like macrolet)
;				   (result 'X))
;			       (dolist (acc accessor)
;				 (setf result `(list (quote ,acc) ,result)))
;			       `(,name ,(list 'X) ,result)))
;			 names)))
;  `(macrolet ,macros . ,forms))
;) 

(defmacro with-cons-places (forest &body body)

  "Locally defines accessors via flet to the leafs of each car/cdr tree defined in the 
specified forest - a list of lists. Each list in the specified forest defines one tree. 
So it is possible to define a hierarchy of accessors to non-leaf-nodes.

Each non-nil symbol acc which occurs in one tree lst denotes an accessor function acc
such that 

 (eq (acc lst) 'acc) is true.

Keyword symbols will be ignored. So if one specifies ((a b) . :ignore) as a tree 
only the accessors a (with #'a = #'caar and #(setf a) = #'(setf caar)) 
and b (with #'b = #'cadar and #'(setf b) = #'(setf cadar)) will be defined.

Examples:

  (with-cons-places ((search-key . attached-value))
    (let ((data '((:a-key       . :a-value) 
                  (:another-key . :another-value)
                  (:some-key     . :old-value))))
      (setf (attached-value (find :some-key data :key #'search-key))
            :new-value )
      (attached-value (find :some-key data :key #'search-key))))
  ==>
  :new-value

  (with-cons-places (( row-col  . attached-value) 
                     ((row col) . :dont-ignore)) ; will be ignored anyway
    (let ((matrix '( ((1 1) . :a11) ((1 2) . :a12) ((1 3) . :a13)
                                    ((2 2) . :a22) 
                     ((3 1) . :a31)                ((3 3) . :a33))))
       (flet ((row-n  (mat n) (remove n mat
                                     :key #'row
                                     :test (complement #'eql)))
              (col-m  (mat m) (remove m mat
                                     :key #'col
                                     :test (complement #'eql))))
         (col-m matrix 2)))) 
   ==>
   (((1 2) . :A12) ((2 2) . :A22))

  (with-cons-places ((() () () () () () () () () () eleventh twelvth thirteenth))
     (eleventh '(1 2 3 4 5 6 7 8 9 10 12)))
  ==> 12
 
  (let ((lst (list 1 2 3 4 5 6 7 8 9 10 12)))
    (with-cons-places ((() () () () () () () () () () eleventh twelvth thirteenth))
       (setf (eleventh lst) 11)
       lst ))
  ==> (1 2 3 4 5 6 7 8 9 10 11)
"

  (labels ((cons-places (tree &optional (result ()))

	     (cond ((null tree)     ())
		   ((keywordp tree) ())	  ; ignore keywords
		   ((symbolp tree)  (list (cons tree ; well, actually, its leaf.
						(reverse result))))
		   ((atom tree)     ())	; this is an error.
		   (T
		    (append 
		     (cons-places (car tree) (cons 'car result))
		     (cons-places (cdr tree) (cons 'cdr result))))))


	   ;; concatenates sequences (cxr cyr czr) into czyxr or fourth, ..., tenth
	   ;; note the inversion in order from (.x. .y. .z.) to .zyx. 
	   (pack-cons-place (lst &aux (pos 0))
	     (cond ((null lst) nil)
		   ((null (cdr lst)) lst)
		   ((null (cddr lst))
		    (cond ((equal lst '(car car)) '(caar))
			  ((equal lst '(cdr car)) '(cadr))
			  ((equal lst '(car cdr)) '(cdar))
			  ((equal lst '(cdr cdr)) '(cddr))
			  (T lst)))	; be safe
		   ((null (cdddr lst))
		    (cond ((equal lst '(car car car)) '(caaar))
			  ((equal lst '(car cdr car)) '(cadar))
			  ((equal lst '(car car cdr)) '(cdaar))
			  ((equal lst '(car cdr cdr)) '(cddar))
			  ((equal lst '(cdr car car)) '(caadr))
			  ((equal lst '(cdr cdr car)) '(caddr))
			  ((equal lst '(cdr car cdr)) '(cdadr))
			  ((equal lst '(cdr cdr cdr)) '(cdddr))
			  ( T lst)))	; be safe
		   ((null (cddddr lst))
		    (cond ((equal lst '(car car car car)) '(caaaar))
			  ((equal lst '(car car cdr car)) '(cadaar))
			  ((equal lst '(car car car cdr)) '(cdaaar))
			  ((equal lst '(car car cdr cdr)) '(cddaar))
			  ((equal lst '(car cdr car car)) '(caadar))
			  ((equal lst '(car cdr cdr car)) '(caddar))
			  ((equal lst '(car cdr car cdr)) '(cdadar))
			  ((equal lst '(car cdr cdr cdr)) '(cdddar))
			  ((equal lst '(cdr car car car)) '(caaadr))
			  ((equal lst '(cdr car cdr car)) '(cadadr))
			  ((equal lst '(cdr car car cdr)) '(cdaadr))
			  ((equal lst '(cdr car cdr cdr)) '(cddadr))
			  ((equal lst '(cdr cdr car car)) '(caaddr))
			  ((equal lst '(cdr cdr cdr car)) '(cadddr))
			  ((equal lst '(cdr cdr car cdr)) '(cdaddr))
			  ((equal lst '(cdr cdr cdr cdr)) '(cddddr))
			  ( T lst)))	; be safe

		   ;; first to tenth are applicable.
		   ;;  Note: first to fourth are already handled by the previous cases:
		   ;;    car, cadr, caddr, or cadddr.
		   ((< (setf pos (position 'car lst)) ; relies on the fact, that only car and cdr exists in lst
		       9)
		    (cons (nth pos '(first second third fourth fifth sixth seventh eighth ninth tenth))
			  (pack-cons-place (nthcdr pos (cdr lst)))))

		   ;; split up the sequence into subsequences of length at most 4.
		   (T (append (pack-cons-place (list (car lst) (cadr lst) (caddr lst) (cadddr lst)))
			      (pack-cons-place (cddddr lst)))))))

    (let* ((accessors (mapcan #'cons-places forest))
	   (names     (mapcar #'car accessors))
	   (functions (mapcan #'(lambda (name-acc)
				  (let ((name (car name-acc))
					(accessor 
					 (pack-cons-place
					  (cdr name-acc)))  
					(result 'X))
				    (dolist (acc accessor)
				      (setf result `(,acc ,result)))
				    `((,name          (X) ,result) ; the reader
				      ((setf ,name) (Y X) (setf ,result Y))))) ; the writer
			      accessors)))
      `(flet ,functions 
	 (declare (ignorable ,@(mapcan #'(lambda (name) 
					   `((function ,name) (function (setf ,name))))
				       names)))
	 . ,body))))




;; efficient binary list length comparison:
;; TODO: (defun list-length<op> (a &rest others) ... )

(defun list-length< (a b)
  (do ((a a (cdr a))
       (b b (cdr b)))
      ((or (null a) (null b)) (and (null a) b))))

(defun list-length> (b a) 
  (do ((a a (cdr a))
       (b b (cdr b)))
      ((or (null a) (null b)) (and (null a) b))))

(defun list-length= (a b)
  (do ((a a (cdr a))
       (b b (cdr b)))
      ((or (null a) (null b)) (and (null a) (null b)))))


(defun list-length<= (a b)
  (do ((a a (cdr a))
       (b b (cdr b)))
      ((or (null a) (null b)) (null a))))

(defun list-length>= (b a)
  (do ((a a (cdr a))
       (b b (cdr b)))
      ((or (null a) (null b)) (null a))))


;;; Currying:
;;; Standard from left:
(defun curry (fn &rest args)
  #'(lambda (&rest argz)
      (apply fn (append args argz))))

;;; from right:
(defun rcurry (fn &rest argz)
  #'(lambda (&rest args)
      (apply fn (append args argz))))


;;; Currying macros:
(defmacro mc-curry (op &rest forms)
  (let ((argssym  (gensym "ARGS")))
     `#'(lambda (&rest ,argssym)
	 (apply ,op ,@forms ,argssym))))

;;; from right:
(defmacro mc-rcurry (op &rest forms)
  (let ((argssym  (gensym "ARGS")))
    `#'(lambda (&rest ,argssym)
	(apply ,op (append ,argssym (list ,@forms))))))


;; this one is missing in CL:
(defun supsetp (list-1 list-2 &key key test test-not)
  (subsetp list-2 list-1 :test (make-test key test test-not)))


(defmacro flet-function-symbols (function-symbols &rest body)
  "flet-function-symbols function-sybols &rest body

Generates an flet environment where each specified symbol 
in function-symbols is taken as the name of a local function 
which applies the variable-assignment of that symbol 
to the agruments.

In short, it generates an flet where (f args) is a shorthand 
for (apply f args) for each symbol f in function symbols."

  `(flet ,(mapcar #'(lambda (fsym)
		      (let ((args (gensym)))
		      `(,fsym (&rest ,args) (apply ,fsym ,args))))
		  function-symbols)
     (declare (ignorable ,@function-symbols)
	      (inline ,@function-symbols))
     ,@body))

(defmacro with-function-wrappers (names &body body) ;http://paste.lisp.org/+2ET3
  "Names is a list ((new-name old-name) *).
The local function new-name can be used  instead of old-name within the body.

New-name may be a different function object than old-name. It only behaves equal.
In this implementation new-name is rather a wrapper than an alias."
 
  `(flet ,(mapcar #'(lambda (name-pair &aux (params (gensym "ARGS")))
		      `(,(car name-pair) (&rest ,params) 
			 (apply (function ,(cadr name-pair)) ,params)))
		  names)
     (declare (inline ,@(mapcar #'car names))
	      (ignorable ,@(mapcar #'car names)))
     ,@body))

(defmacro with-function-aliases (names &body body) ;http://paste.lisp.org/+2ET3
  "Names is a list ((new-name old-name) *).
The local macro new-name can be used instead of old-name within the body.

New-name is a macro which will be expanded into old-name so it cannot be used as 
a function designator."
 
  `(macrolet ,(mapcar #'(lambda (name-pair &aux (params (gensym "ARGS")))
		      `(,(car name-pair) (&body ,params) 
			 (cons ',(cadr name-pair) ,params)))
		  names)
     ,@body))


(defmacro with-symbol-aliases (names &body body)
  "Names is a list ((used-name name-meant) *).
Replaces every occurrence of used-name by name-meant in parallel."

  (let* ((tmpnames  (mapcar #'(lambda (used-meant) 
				(cons (gensym "TMP") used-meant))
			    names))
	 (result (copy-tree body)))
    (dolist (tmp-used-meant tmpnames)
      (setf result
	    (nsubst (car tmp-used-meant) (cadr tmp-used-meant)
		    result
		   :test #'eq)))

    (dolist (tmp-used-meant tmpnames (cons 'progn result))
      (setf result
	    (nsubst (caddr tmp-used-meant) (car tmp-used-meant)
		    result
		   :test #'eq)))))

(defmacro with-symbols-aliases* (names &body body)
  "Names is a list ((used-name name-meant) *).
Sequencially replaces each occurrence of used-name by name-meant."

  (let* ((result (copy-tree body)))
    (dolist (used-meant names (cons 'progn result))
      (setf result
	    (nsubst (cadr used-meant) (car used-meant)
		    result
		   :test #'eq)))))


(defun remove-nth (n list)
  (if (zerop n) 
      (cdr list)
      (let ((head ())
	    (tail list))
	(dotimes (i n)
	  (push (pop tail) head))
	(nreconc head (cdr tail)))))


(defun delete-nth (n list)
  (if (zerop n) 
      (cdr list)
      (let ((tail (nthcdr (1- n) list)))
	(rplacd tail (cddr tail))
	list )))


(defmacro long-and (&rest body)
  "Behaves like and except that it evaluates every given expression."
  (let ((result (gensym)))
    `(let ((,result T))
       ,@(mapcar #'(lambda (expr)
		   `(setf ,result (and ,expr ,result)))
	       body)
       ,result)))

(defmacro long-or (&rest body)
  "Behaves like or except that it evaluates every given expression."
  (let ((result (gensym)))
    `(let ((,result nil))
       ,@(mapcar #'(lambda (expr)
		   `(setf ,result (or ,expr ,result)))
	       body)
       ,result)))


(defmacro defun-with-default-output-keys (name args &body body)
  "Extends a defun by the output control keyword parameters of write with 
default init-values of the corresponding output control variables and binds 
these variables to the corresponding values during the execution of body.

See CLHS write for a list of these variables. 
The stream keyword will NOT be set.

An &key parameter keyword must be specified. If one of the output control keywords 
is explicitly specified it will not be defined again."
  (let* ((keywords        (member '&key args))
	 (before-keywords (nreverse (nthcdr (length keywords) (reverse args))))
	 (after-keywords (member-if #'(lambda (arg) (member arg '(&allow-other-keys &aux &body &environment #|&key|# &optional &rest &whole)))
				    keywords))
	 (new-keywords   `((:array            ,(gensym) *print-array*)
			   (:base             ,(gensym) *print-base* )                   
			   (:case             ,(gensym) *print-case* )                   
			   (:circle           ,(gensym) *print-circle*)                  
			   (:escape           ,(gensym) *print-escape*)                  
			   (:gensym           ,(gensym) *print-gensym*)                  
			   (:length           ,(gensym) *print-length*)                 
			   (:level            ,(gensym) *print-level* )                  
			   (:lines            ,(gensym) *print-lines* )                  
			   (:miser-width      ,(gensym) *print-miser-width*)             
			   (:pprint-dispatch  ,(gensym) *print-pprint-dispatch*)         
			   (:pretty           ,(gensym) *print-pretty*)                  
			   (:radix            ,(gensym) *print-radix* )                  
			   (:readably         ,(gensym) *print-readably*)                
			   (:right-margin     ,(gensym) *print-right-margin*)))
	 (declarations   (let ((result ()))
			   (dolist (form body (nreverse result))
			     (if (and (consp form)
				      (eq (car form) 'declare))
				 (push form result)
				 (return (nreverse result))))))
	 (documentation (let ((doc (nth (length declarations) body)))
			  (if (stringp doc) (list doc) ())))
	 (forms         (nthcdr (+ (length declarations) (length documentation)) body)))
    (unless keywords
      (error "There must be a (possibly empty) keyword section &key in the argument list ~S." args))
    (setf keywords (nreverse (nthcdr (length after-keywords) (reverse keywords))))
    (dolist (key (cdr keywords))
      (let* ((key-name-var (if (symbolp key) key (car key)))
	     (key-name     (if (symbolp key-name-var) key-name-var (car key-name-var)))
	     (dropee (assoc (symbol-name key-name) new-keywords :test #'string= :key #'symbol-name)))
	(when dropee
	  (setf new-keywords (remove dropee new-keywords :test #'string=)))))
    `(defun ,name (,@before-keywords 
		   ,@keywords 
		   ,@(mapcar #'(lambda (key-var-svar) `((,(car key-var-svar) ,(cadr key-var-svar)) ,(caddr key-var-svar)))
			     new-keywords) 
		   ,@after-keywords)
       ,@declarations
       ,@(if forms documentation ())
       (let ,(mapcar #'(lambda (key-var-svar) 
			 `(,(caddr key-var-svar) ,(cadr key-var-svar)))
		     new-keywords)
	 ,@(if forms forms documentation)))))


(defun project (i &rest args)
  (nth i args))


(defmacro multiple-value-let (bindings &body body)
  (let ((mv-bindings ())
	(sv-bindings ()))
    (dolist (binding bindings)
      (cond ((and (not (atom binding)) (null (car binding)))
	     (push (cons () binding) mv-bindings))
	    ((or (atom binding) (atom (car binding)))
	     (push binding sv-bindings))
	    (t
	     (push (cons (mapcar #'(lambda (x) (declare (ignore x)) (gensym "TMP")) 
				 (car binding)) 
			 binding)
		   mv-bindings))))
    (setf mv-bindings (nreverse mv-bindings)
	  sv-bindings (nreverse sv-bindings))
    (cond ((null mv-bindings)
	   `(let ,bindings ,@body))
	  ((null sv-bindings)
	   `(let ,(reduce #'append ; NOT mapcan!!!
			  (mapcar #'car mv-bindings)) ; protect the gensymed vars.
	      ;; set the gensymed vars:
	      ,@(mapcan #'(lambda (binding)
			    (cond ((null (caddr binding)) ())
				  ((null (car binding))
				   (list (caddr binding)))
				  ((null (cdar binding))
				   `((setq ,(caar binding) ,(caddr binding))))
				  (T
				   `((multiple-value-setq ,(car binding) ,(caddr binding))))))
			mv-bindings)
	      ;; bind the user vars
	      (let ,(mapcan #'(lambda (binding) 
				(mapcar #'list 
					(cadr binding) ; the user var name
					(car  binding))) ; the bound gensym
			    mv-bindings)
		,@body)))
	  (T
	   `(multiple-value-let 
	     ,(mapcar #'(lambda (binding)
			(cond ((atom binding)
			       `((,binding) ()))
			      ((atom (car binding))
			       `((,(car binding)) ,@(cdr binding)))
			      (T binding)))
		    bindings)
	     ,@body)))))


;(defun randomize (lst)
;  (let ((lst    (copy-list lst))
;	(n      (length lst))
;	(result ()))
;    (dotimes (i n result)
;      (let ((rnd (random n)))
;	(if (zerop rnd) 
;	    (push (pop lst) result)
;	    (let ((tail (nthcdr (1- rnd) lst)))   ; (a_0 a_1 ... . (a_rnd-1 !takeme! b c ...))
;	      (push (cadr tail) result)           ; (a_0 a_1 ... . (a_rnd-1 !takeme! b c ...))
;	      (setf (cdr tail) (cddr tail)))))    ; (a_0 a_1 ... . (a b c ...))
;      (decf n))))


(defun intersectionp (list-1 list-2 &key key (test #'eql) test-not) ; => generalized boolean

  "Returns the same generalized boolean value as (intersection list-1 list-2) but 
terminates as soon as the first element in (intersection list-1 list-2) is found
\(i.e. until the first pair of elements from either list satisfies the given test)."

  (if test-not
    (intersectionp list-1 list-2 :key key :test (complement test-not))
    (progn 
      (intersection list-1 list-2 :key key 
		    :test #'(lambda (arg-1 arg-2)
			      (when (funcall test arg-1 arg-2)
				(return-from intersectionp T))))
      nil)))


(defun reduced (fxn &key key from-end (start 0) end (initial-value () initial-value-supplied))

  "Extends a binary function to an n-ary function as if by calling reduce:

\(funcall (reduced fx :key-1 key-1 ...) arg1 arg2 ... argn) 
 == (reduce fx (list arg-1 arg-2 ... arg-n) :key-1 key-1 ...)."

  (if initial-value-supplied
      #'(lambda (&rest args)
	  (reduce fxn args :key key :from-end from-end :start start :end end :initial-value initial-value))
      #'(lambda (&rest args)
	  (reduce fxn args :key key :from-end from-end :start start :end end))))


(defmacro obsolete (function &optional alternative)
  (if alternative 
      `(define-compiler-macro ,function (&whole form &rest args)
	 (declare (ignore args))
	 (warn 'style-warning "~S is declared obsolete. Use ~S instead." ',function ',alternative)
	 form)
      `(define-compiler-macro ,function (&whole form &rest args)
	 (declare (ignore args))
      	 (warn 'style-warning "~S is declared obsolete." ',function)
	 form)))
